"""Support for optimizing tensor networks using automatic differentiation to
automatically derive gradients for input to scipy optimizers.
"""
import re
import warnings
import functools
import importlib
from collections.abc import Iterable

import tqdm
import numpy as np
from autoray import to_numpy, astype, get_dtype_name

from .tensor_core import (
    PTensor,
    tags_to_oset,
)
from .contraction import contract_backend
from .interface import get_jax
from ..core import prod
from ..utils import ensure_dict, tree_map, tree_flatten, tree_unflatten

if importlib.util.find_spec("jax") is not None:
    _DEFAULT_BACKEND = "jax"
elif importlib.util.find_spec("tensorflow") is not None:
    _DEFAULT_BACKEND = "tensorflow"
elif importlib.util.find_spec("torch") is not None:
    _DEFAULT_BACKEND = "torch"
else:
    _DEFAULT_BACKEND = "autograd"


_REAL_CONVERSION = {
    "float32": "float32",
    "float64": "float64",
    "complex64": "float32",
    "complex128": "float64",
}

_COMPLEX_CONVERSION = {
    "float32": "complex64",
    "float64": "complex128",
    "complex64": "complex64",
    "complex128": "complex128",
}


class ArrayInfo:
    """Simple container for recording size and dtype information about arrays."""

    __slots__ = (
        "shape",
        "size",
        "dtype",
        "iscomplex",
        "real_size",
        "equivalent_real_type",
        "equivalent_complex_type",
    )

    def __init__(self, array):
        self.shape = array.shape
        self.size = prod(self.shape)
        self.dtype = get_dtype_name(array)
        self.equivalent_real_type = _REAL_CONVERSION[self.dtype]
        self.equivalent_complex_type = _COMPLEX_CONVERSION[self.dtype]
        self.iscomplex = "complex" in self.dtype
        self.real_size = self.size * (2 if self.iscomplex else 1)

    def __repr__(self):
        return (
            "ArrayInfo("
            f"shape={self.shape}, "
            f"size={self.size}, "
            f"dtype={self.dtype}"
            ")"
        )


class Vectorizer:
    """Object for mapping back and forth between any pytree of mixed
    real/complex n-dimensional arrays to a single, real, double precision numpy
    vector, as required by ``scipy.optimize`` routines.

    Parameters
    ----------
    tree : pytree of array
        Any nested container of arrays, which will be flattened and packed into
        a single float64 vector.
    is_leaf : callable, optional
        A function which takes a single argument and returns ``True`` if it is
        a leaf node in the tree and should be extracted, ``False`` otherwise.
        Defaults to everything that is not a tuple, list or dict.
    """

    def __init__(self, tree):
        arrays = []
        self.infos = []
        self.d = 0

        def extracter(x):
            arrays.append(x)
            info = ArrayInfo(x)
            self.infos.append(info)
            self.d += info.real_size
            return info

        self.ref_tree = tree_map(extracter, tree)
        self.pack(arrays)

    def pack(self, tree, name="vector"):
        """Take ``arrays`` and pack their values into attribute `.{name}`, by
        default `.vector`.
        """
        arrays = tree_flatten(tree)

        # create the vector if it doesn't exist yet
        if not hasattr(self, name):
            setattr(self, name, np.empty(self.d, "float64"))
        x = getattr(self, name)

        i = 0
        for array, info in zip(arrays, self.infos):
            if not isinstance(array, np.ndarray):
                array = to_numpy(array)
            # flatten
            if info.iscomplex:
                # view as real array of double the length
                real_view = array.reshape(-1).view(info.equivalent_real_type)
            else:
                real_view = array.reshape(-1)
            # pack into our vector
            f = i + info.real_size
            x[i:f] = real_view
            i = f

        return x

    def unpack(self, vector=None):
        """Turn the single, flat ``vector`` into a sequence of arrays."""
        if vector is None:
            vector = self.vector

        i = 0
        arrays = []
        for info in self.infos:
            # get the linear slice
            f = i + info.real_size
            array = vector[i:f]
            i = f
            if info.iscomplex:
                # view as complex array of half the length
                array = array.view(np.complex128)
            # reshape (inplace)
            array.shape = info.shape
            if get_dtype_name(array) != info.dtype:
                # cast as original dtype
                array = astype(array, info.dtype)
            arrays.append(array)

        return tree_unflatten(
            arrays, self.ref_tree, lambda x: isinstance(x, ArrayInfo)
        )


_VARIABLE_TAG = "__VARIABLE{}__"
variable_finder = re.compile(r"__VARIABLE(\d+)__")


def _parse_opt_in(tn, tags, shared_tags, to_constant):
    """Parse a tensor network where tensors are assumed to be constant unless
    tagged.
    """
    tn_ag = tn.copy()
    variables = []

    # tags where each individual tensor should get a separate variable
    individual_tags = tags - shared_tags

    # handle tagged tensors that are not shared
    for t in tn_ag.select_tensors(individual_tags, "any"):
        # append the raw data but mark the corresponding tensor
        # for reinsertion
        data = t.get_params()
        variables.append(data)
        t.add_tag(_VARIABLE_TAG.format(len(variables) - 1))

    # handle shared tags
    for tag in shared_tags:

        var_name = _VARIABLE_TAG.format(len(variables))
        test_data = None

        for t in tn_ag.select_tensors(tag):
            data = t.get_params()

            # detect that this tensor is already variable tagged and skip
            # if it is
            if any(variable_finder.match(tag) for tag in t.tags):
                warnings.warn(
                    "TNOptimizer warning, tensor tagged with"
                    " multiple `tags` or `shared_tags`."
                )
                continue

            if test_data is None:
                # create variable and store data
                variables.append(data)
                test_data = data
            else:
                # check that the shape of the variable's data matches the
                # data of this new tensor
                if test_data.shape != data.shape:
                    raise ValueError(
                        "TNOptimizer error, a `shared_tags` tag "
                        "covers tensors with different numbers of"
                        " params."
                    )

            # mark the corresponding tensor for reinsertion
            t.add_tag(var_name)

    # iterate over tensors which *don't* have any of the given tags
    for t in tn_ag.select_tensors(tags, which="!any"):
        t.modify(apply=to_constant)

    return tn_ag, variables


def _parse_opt_out(
    tn,
    constant_tags,
    to_constant,
):
    """Parse a tensor network where tensors are assumed to be variables unless
    tagged.
    """
    tn_ag = tn.copy()
    variables = []

    for t in tn_ag:

        if t.tags & constant_tags:
            t.modify(apply=to_constant)
            continue

        # append the raw data but mark the corresponding tensor
        # for reinsertion
        data = t.get_params()
        variables.append(data)
        t.add_tag(_VARIABLE_TAG.format(len(variables) - 1))

    return tn_ag, variables


def parse_network_to_backend(
    tn,
    to_constant,
    tags=None,
    shared_tags=None,
    constant_tags=None,
):
    """
    Parse tensor network to:

        - identify the dimension of the optimisation space and the initial
          point of the optimisation from the current values in the tensor
          network,
        - add variable tags to individual tensors so that optimisation vector
          values can be efficiently reinserted into the tensor network.

    There are two different modes:

        - 'opt in' : `tags` (and optionally `shared_tags`) are specified and
          only these tensor tags will be optimised over. In this case
          `constant_tags` is ignored if it is passed,
        - 'opt out' : `tags` is not specified. In this case all tensors will be
          optimised over, unless they have one of `constant_tags` tags.

    Parameters
    ----------
    tn : TensorNetwork
        The initial tensor network to parse.
    to_constant : Callable
        Function that fixes a tensor as constant.
    tags : str, or sequence of str, optional
        Set of opt-in tags to optimise.
    shared_tags : str, or sequence of str, optional
        Subset of opt-in tags to joint optimise i.e. all tensors with tag s in
        shared_tags will correspond to the same optimisation variables.
    constant_tags : str, or sequence of str, optional
        Set of opt-out tags if `tags` not passed.

    Returns
    -------
    tn_ag : TensorNetwork
        Tensor network tagged for reinsertion of optimisation variable values.
    variables : list
        List of variables extracted from ``tn``.
    """
    tags = tags_to_oset(tags)
    shared_tags = tags_to_oset(shared_tags)
    constant_tags = tags_to_oset(constant_tags)

    if tags | shared_tags:
        # opt_in
        if not (tags & shared_tags) == shared_tags:
            tags = tags | shared_tags
            warnings.warn(
                "TNOptimizer warning, some `shared_tags` are missing"
                " from `tags`. Automatically adding these missing"
                " `shared_tags` to `tags`."
            )
        if constant_tags:
            warnings.warn(
                "TNOptimizer warning, if `tags` or `shared_tags` are"
                " specified then `constant_tags` is ignored - "
                "consider instead untagging those tensors."
            )
        return _parse_opt_in(
            tn,
            tags,
            shared_tags,
            to_constant,
        )

    # opt-out
    return _parse_opt_out(
        tn,
        constant_tags,
        to_constant,
    )


def convert_raw_arrays(x, f):
    """Given a ``TensorNetwork``, ``Tensor``, or other possibly structured raw
    array, return a copy where the underyling data has had ``f``
    applied to it. Structured raw arrays should implement the
    ``tree = get_params()`` and ``set_params(tree)`` methods which get or set
    their underlying data using an arbitrary pytree.
    """
    try:
        # Tensor, TensorNetwork...
        x = x.copy()
        x.apply_to_arrays(f)
        return x
    except AttributeError:
        pass

    try:
        # raw structured arrays that provide the {get|set}_params interface
        x = x.copy()
        x.set_params(tree_map(f, x.get_params()))
        return x
    except AttributeError:
        pass

    # other raw arrays
    return f(x)


@functools.lru_cache(1)
def get_autograd():
    import autograd

    return autograd


class AutoGradHandler:
    def __init__(self, device="cpu"):
        if device != "cpu":
            raise ValueError(
                "`autograd` currently is only " "backed by cpu, numpy arrays."
            )

    def to_variable(self, x):
        return np.asarray(x)

    def to_constant(self, x):
        return np.asarray(x)

    def setup_fn(self, fn):
        autograd = get_autograd()
        self._backend_fn = fn
        self._value_and_grad = autograd.value_and_grad(fn)
        self._hvp = autograd.hessian_vector_product(fn)

    def value(self, arrays):
        return self._backend_fn(arrays)

    def value_and_grad(self, arrays):
        loss, grads = self._value_and_grad(arrays)
        return loss, tree_map(lambda x: x.conj(), grads)


class JaxHandler:
    def __init__(self, jit_fn=True, device=None):
        self.jit_fn = jit_fn
        self.device = device

    def to_variable(self, x):
        jax = get_jax()
        return jax.numpy.asarray(x)

    def to_constant(self, x):
        jax = get_jax()
        return jax.numpy.asarray(x)

    def setup_fn(self, fn):
        jax = get_jax()
        if self.jit_fn:
            self._backend_fn = jax.jit(fn, backend=self.device)
            self._value_and_grad = jax.jit(
                jax.value_and_grad(fn), backend=self.device
            )
        else:
            self._backend_fn = fn
            self._value_and_grad = jax.value_and_grad(fn)

        self._setup_hessp(fn)

    def _setup_hessp(self, fn):
        jax = get_jax()

        def hvp(primals, tangents):
            return jax.jvp(jax.grad(fn), (primals,), (tangents,))[1]

        if self.jit_fn:
            hvp = jax.jit(hvp, device=self.device)

        self._hvp = hvp

    def value(self, arrays):
        jax_arrays = tree_map(self.to_constant, arrays)
        return to_numpy(self._backend_fn(jax_arrays))

    def value_and_grad(self, arrays):
        loss, grads = self._value_and_grad(arrays)
        return (
            loss,
            tree_map(lambda x: to_numpy(x.conj()), grads),
        )

    def hessp(self, primals, tangents):
        jax_arrays = self._hvp(primals, tangents)
        return tree_map(to_numpy, jax_arrays)


@functools.lru_cache(1)
def get_tensorflow():
    import tensorflow

    return tensorflow


class TensorFlowHandler:
    def __init__(
        self,
        jit_fn=False,
        autograph=False,
        experimental_compile=False,
        device=None,
    ):
        self.jit_fn = jit_fn
        self.autograph = autograph
        self.experimental_compile = experimental_compile
        self.device = device

    def to_variable(self, x):
        tf = get_tensorflow()
        if self.device is None:
            return tf.Variable(x)
        with tf.device(self.device):
            return tf.Variable(x)

    def to_constant(self, x):
        tf = get_tensorflow()
        if self.device is None:
            return tf.constant(x)
        with tf.device(self.device):
            return tf.constant(x)

    def setup_fn(self, fn):
        tf = get_tensorflow()
        if self.jit_fn:
            self._backend_fn = tf.function(
                fn,
                autograph=self.autograph,
                experimental_compile=self.experimental_compile,
            )
        else:
            self._backend_fn = fn

    def value(self, arrays):
        tf_arrays = tree_map(self.to_constant, arrays)
        return to_numpy(self._backend_fn(tf_arrays))

    def value_and_grad(self, arrays):
        tf = get_tensorflow()

        variables = tree_map(self.to_variable, arrays)

        with tf.GradientTape() as t:
            result = self._backend_fn(variables)

        tf_grads = t.gradient(
            result,
            variables,
            # want to return zeros for unconnected gradients
            unconnected_gradients=tf.UnconnectedGradients.ZERO,
        )

        grads = tree_map(to_numpy, tf_grads)
        loss = to_numpy(result)
        return loss, grads


@functools.lru_cache(1)
def get_torch():
    import torch

    return torch


class TorchHandler:
    def __init__(self, jit_fn=False, device=None):
        torch = get_torch()
        self.jit_fn = jit_fn
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device

    def to_variable(self, x):
        torch = get_torch()
        return torch.tensor(x).to(self.device).requires_grad_()

    def to_constant(self, x):
        torch = get_torch()
        return torch.tensor(x).to(self.device)

    def setup_fn(self, fn):
        self._fn = fn
        self._backend_fn = None

    def _setup_backend_fn(self, arrays):
        torch = get_torch()
        if self.jit_fn:
            example_inputs = (tree_map(self.to_variable, arrays),)
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action='ignore',
                    message=".*can't record the data flow of Python values.*",
                )
                self._backend_fn = torch.jit.trace(
                    self._fn, example_inputs=example_inputs
                )
        else:
            self._backend_fn = self._fn

    def value(self, arrays):
        if self._backend_fn is None:
            self._setup_backend_fn(arrays)
        torch_arrays = tree_map(self.to_constant, arrays)
        return to_numpy(self._backend_fn(torch_arrays))

    def value_and_grad(self, arrays):
        if self._backend_fn is None:
            self._setup_backend_fn(arrays)

        variables = tree_map(self.to_variable, arrays)
        result = self._backend_fn(variables)

        def get_gradient_from_torch(t):
            if t.grad is None:
                return np.zeros(t.shape, dtype=get_dtype_name(t))
            return to_numpy(t.grad).conj()

        result.backward()
        grads = tree_map(get_gradient_from_torch, variables)

        loss = to_numpy(result)
        return loss, grads


_BACKEND_HANDLERS = {
    "numpy": AutoGradHandler,
    "autograd": AutoGradHandler,
    "jax": JaxHandler,
    "tensorflow": TensorFlowHandler,
    "torch": TorchHandler,
}


class MultiLossHandler:
    def __init__(self, autodiff_backend, executor=None, **backend_opts):
        self.autodiff_backend = autodiff_backend
        self.backend_opts = backend_opts
        self.executor = executor

        # start just with one, as we don't don't know how many functions yet
        h0 = _BACKEND_HANDLERS[autodiff_backend](**backend_opts)
        self.handlers = [h0]
        # ... but we do need access to `to_constant`
        self.to_constant = h0.to_constant

    def setup_fn(self, funcs):
        fn0, *fns = funcs
        self.handlers[0].setup_fn(fn0)
        for fn in fns:
            h = _BACKEND_HANDLERS[self.autodiff_backend](**self.backend_opts)
            h.setup_fn(fn)
            self.handlers.append(h)

    def _value_seq(self, arrays):
        return sum(h.value(arrays) for h in self.handlers)

    def _value_par_seq(self, arrays):
        futures = [
            self.executor.submit(h.value, arrays) for h in self.handlers
        ]
        return sum(f.result() for f in futures)

    def value(self, arrays):
        if self.executor is not None:
            return self._value_par(arrays)
        return self._value_seq(arrays)

    def _value_and_grad_seq(self, arrays):
        h0, *hs = self.handlers
        loss, grads = h0.value_and_grad(arrays)
        # need to make arrays writeable for efficient inplace sum
        grads = list(map(np.array, grads))
        for h in hs:
            loss_i, grads_i = h.value_and_grad(arrays)
            loss = loss + loss_i
            for i, g_i in enumerate(grads_i):
                grads[i] += g_i
        return loss, grads

    def _value_and_grad_par(self, arrays):
        futures = [
            self.executor.submit(h.value_and_grad, arrays)
            for h in self.handlers
        ]
        results = (f.result() for f in futures)

        # get first result
        loss, grads = next(results)
        grads = list(map(np.array, grads))

        # process remaining results
        for loss_i, grads_i in results:
            loss = loss + loss_i
            for i, g_i in enumerate(grads_i):
                grads[i] += g_i

        return loss, grads

    def value_and_grad(self, arrays):
        if self.executor is not None:
            return self._value_and_grad_par(arrays)
        return self._value_and_grad_seq(arrays)


def inject_(arrays, tn):
    for t in tn:
        for tag in t.tags:
            match = variable_finder.match(tag)
            if match is not None:
                i = int(match.groups(1)[0])
                t.set_params(arrays[i])
                break


class SGD:
    """Stateful ``scipy.optimize.minimize`` compatible implementation of
    stochastic gradient descent with momentum.

    Adapted from ``autograd/misc/optimizers.py``.
    """

    def __init__(self):
        from scipy.optimize import OptimizeResult

        self.OptimizeResult = OptimizeResult
        self._i = 0
        self._velocity = None

    def get_velocity(self, x):
        if self._velocity is None:
            self._velocity = np.zeros_like(x)
        return self._velocity

    def __call__(
        self,
        fun,
        x0,
        jac,
        args=(),
        learning_rate=0.1,
        mass=0.9,
        maxiter=1000,
        callback=None,
        bounds=None,
        **kwargs,
    ):

        x = x0
        velocity = self.get_velocity(x)

        for _ in range(maxiter):
            self._i += 1

            g = jac(x)

            if callback and callback(x):
                break

            velocity = mass * velocity - (1.0 - mass) * g
            x = x + learning_rate * velocity

            if bounds is not None:
                x = np.clip(x, bounds[:, 0], bounds[:, 1])

        # save for restart
        self._velocity = velocity

        return self.OptimizeResult(
            x=x, fun=fun(x), jac=g, nit=self._i, nfev=self._i, success=True
        )


class RMSPROP:
    """Stateful ``scipy.optimize.minimize`` compatible implementation of
    root mean squared prop: See Adagrad paper for details.

    Adapted from ``autograd/misc/optimizers.py``.
    """

    def __init__(self):
        from scipy.optimize import OptimizeResult

        self.OptimizeResult = OptimizeResult
        self._i = 0
        self._avg_sq_grad = None

    def get_avg_sq_grad(self, x):
        if self._avg_sq_grad is None:
            self._avg_sq_grad = np.ones_like(x)
        return self._avg_sq_grad

    def __call__(
        self,
        fun,
        x0,
        jac,
        args=(),
        learning_rate=0.1,
        gamma=0.9,
        eps=1e-8,
        maxiter=1000,
        callback=None,
        bounds=None,
        **kwargs,
    ):
        x = x0
        avg_sq_grad = self.get_avg_sq_grad(x)

        for _ in range(maxiter):
            self._i += 1

            g = jac(x)

            if callback and callback(x):
                break

            avg_sq_grad = avg_sq_grad * gamma + g**2 * (1 - gamma)
            x = x - learning_rate * g / (np.sqrt(avg_sq_grad) + eps)

            if bounds is not None:
                x = np.clip(x, bounds[:, 0], bounds[:, 1])

        # save for restart
        self._avg_sq_grad = avg_sq_grad

        return self.OptimizeResult(
            x=x, fun=fun(x), jac=g, nit=self._i, nfev=self._i, success=True
        )


class ADAM:
    """Stateful ``scipy.optimize.minimize`` compatible implementation of
    ADAM - http://arxiv.org/pdf/1412.6980.pdf.

    Adapted from ``autograd/misc/optimizers.py``.
    """

    def __init__(self):
        from scipy.optimize import OptimizeResult

        self.OptimizeResult = OptimizeResult
        self._i = 0
        self._m = None
        self._v = None

    def get_m(self, x):
        if self._m is None:
            self._m = np.zeros_like(x)
        return self._m

    def get_v(self, x):
        if self._v is None:
            self._v = np.zeros_like(x)
        return self._v

    def __call__(
        self,
        fun,
        x0,
        jac,
        args=(),
        learning_rate=0.001,
        beta1=0.9,
        beta2=0.999,
        eps=1e-8,
        maxiter=1000,
        callback=None,
        bounds=None,
        **kwargs,
    ):
        x = x0
        m = self.get_m(x)
        v = self.get_v(x)

        for _ in range(maxiter):
            self._i += 1

            g = jac(x)

            if callback and callback(x):
                break

            m = (1 - beta1) * g + beta1 * m  # first  moment estimate.
            v = (1 - beta2) * (g**2) + beta2 * v  # second moment estimate.
            mhat = m / (1 - beta1 ** (self._i))  # bias correction.
            vhat = v / (1 - beta2 ** (self._i))
            x = x - learning_rate * mhat / (np.sqrt(vhat) + eps)

            if bounds is not None:
                x = np.clip(x, bounds[:, 0], bounds[:, 1])

        # save for restart
        self._m = m
        self._v = v

        return self.OptimizeResult(
            x=x, fun=fun(x), jac=g, nit=self._i, nfev=self._i, success=True
        )


class NADAM:
    """Stateful ``scipy.optimize.minimize`` compatible implementation of
    NADAM - [Dozat - http://cs229.stanford.edu/proj2015/054_report.pdf].

    Adapted from ``autograd/misc/optimizers.py``.
    """

    def __init__(self):
        from scipy.optimize import OptimizeResult

        self.OptimizeResult = OptimizeResult
        self._i = 0
        self._m = None
        self._v = None
        self._mus = None

    def get_m(self, x):
        if self._m is None:
            self._m = np.zeros_like(x)
        return self._m

    def get_v(self, x):
        if self._v is None:
            self._v = np.zeros_like(x)
        return self._v

    def get_mus(self, beta1):
        if self._mus is None:
            self._mus = [1, beta1 * (1 - 0.5 * 0.96**0.004)]
        return self._mus

    def __call__(
        self,
        fun,
        x0,
        jac,
        args=(),
        learning_rate=0.001,
        beta1=0.9,
        beta2=0.999,
        eps=1e-8,
        maxiter=1000,
        callback=None,
        bounds=None,
        **kwargs,
    ):
        x = x0
        m = self.get_m(x)
        v = self.get_v(x)
        mus = self.get_mus(beta1)

        for _ in range(maxiter):
            self._i += 1

            # this is ``mu[t + 1]`` -> already computed ``mu[t]``
            self._mus.append(
                beta1 * (1 - 0.5 * 0.96 ** (0.004 * (self._i + 1)))
            )

            g = jac(x)

            if callback and callback(x):
                break

            gd = g / (1 - np.prod(self._mus[:-1]))
            m = beta1 * m + (1 - beta1) * g
            md = m / (1 - np.prod(self._mus))
            v = beta2 * v + (1 - beta2) * g**2
            vd = v / (1 - beta2**self._i)
            mhat = (1 - self._mus[self._i]) * gd + self._mus[self._i + 1] * md

            x = x - learning_rate * mhat / (np.sqrt(vd) + eps)

            if bounds is not None:
                x = np.clip(x, bounds[:, 0], bounds[:, 1])

        # save for restart
        self._m = m
        self._v = v
        self._mus = mus

        return self.OptimizeResult(
            x=x, fun=fun(x), jac=g, nit=self._i, nfev=self._i, success=True
        )


_STOC_GRAD_METHODS = {
    "sgd": SGD,
    "rmsprop": RMSPROP,
    "adam": ADAM,
    "nadam": NADAM,
}


class MakeArrayFn:
    """Class wrapper so picklable."""

    __name__ = "MakeArrayFn"

    def __init__(self, tn_opt, loss_fn, norm_fn, autodiff_backend):
        self.tn_opt = tn_opt
        self.loss_fn = loss_fn
        self.norm_fn = norm_fn
        self.autodiff_backend = autodiff_backend

    def __call__(self, arrays):
        # copy the TN so norm and loss functions can modify in place
        # XXX: make optional for efficiency?
        tn_compute = self.tn_opt.copy()
        inject_(arrays, tn_compute)

        # set backend explicitly as maybe mixing with numpy arrays
        with contract_backend(self.autodiff_backend):
            return self.loss_fn(self.norm_fn(tn_compute))


def identity_fn(x):
    return x


class TNOptimizer:
    """Globally optimize tensors within a tensor network with respect to any
    loss function via automatic differentiation. If parametrized tensors are
    used, optimize the parameters rather than the raw arrays.

    Parameters
    ----------
    tn : TensorNetwork
        The core tensor network structure within which to optimize tensors.
    loss_fn : callable or sequence of callable
        The function that takes ``tn`` (as well as ``loss_constants`` and
        ``loss_kwargs``) and returns a single real 'loss' to be minimized.
        For Hamiltonians which can be represented as a sum over terms, an
        iterable collection of terms (e.g. list) can be given instead. In that
        case each term is evaluated independently and the sum taken as loss_fn.
        This can reduce the total memory requirements or allow for
        parallelization (see ``executor``).
    norm_fn : callable, optional
        A function to call before ``loss_fn`` that prepares or 'normalizes' the
        raw tensor network in some way.
    loss_constants : dict, optional
        Extra tensor networks, tensors, dicts/list/tuples of arrays, or arrays
        which will be supplied to ``loss_fn`` but also converted to the correct
        backend array type.
    loss_kwargs : dict, optional
        Extra options to supply to ``loss_fn`` (unlike ``loss_constants`` these
        are assumed to be simple options that don't need conversion).
    tags : str, or sequence of str, optional
        If supplied, only optimize tensors with any of these tags.
    shared_tags : str, or sequence of str, optional
        If supplied, each tag in ``shared_tags`` corresponds to a group of
        tensors to be optimized together.
    constant_tags : str, or sequence of str, optional
        If supplied, skip optimizing tensors with any of these tags. This
        'opt-out' mode is overridden if either ``tags`` or ``shared_tags`` is
        supplied.
    loss_target : float, optional
        Stop optimizing once this loss value is reached.
    optimizer : str, optional
        Which ``scipy.optimize.minimize`` optimizer to use (the ``'method'``
        kwarg of that function). In addition, ``quimb`` implements a few custom
        optimizers compatible with this interface that you can reference by
        name - ``{'adam', 'nadam', 'rmsprop', 'sgd'}``.
    executor : None or Executor, optional
        To be used with term-by-term Hamiltonians. If supplied, this executor
        is used  to parallelize the evaluation. Otherwise each term is
        evaluated in sequence. It should implement the basic
        concurrent.futures (PEP 3148) interface.
    progbar : bool, optional
        Whether to show live progress.
    bounds : None or (float, float), optional
        Constrain the optimized tensor entries within this range (if the scipy
        optimizer supports it).
    autodiff_backend : {'jax', 'autograd', 'tensorflow', 'torch'}, optional
        Which backend library to use to perform the automatic differentation
        (and computation).
    callback : callable, optional
        A function to call after each optimization step. It should take the
        current ``TNOptimizer`` instance as its only argument. Information such
        as the current loss and number of evaluations can then be accessed::

            def callback(tnopt):
                print(tnopt.nevals, tnopt.loss)

    backend_opts
        Supplied to the backend function compiler and array handler. For
        example ``jit_fn=True`` or ``device='cpu'`` .
    """

    def __init__(
        self,
        tn,
        loss_fn,
        norm_fn=None,
        loss_constants=None,
        loss_kwargs=None,
        tags=None,
        shared_tags=None,
        constant_tags=None,
        loss_target=None,
        optimizer="L-BFGS-B",
        progbar=True,
        bounds=None,
        autodiff_backend="AUTO",
        executor=None,
        callback=None,
        **backend_opts,
    ):
        self.progbar = progbar
        self.tags = tags
        self.shared_tags = shared_tags
        self.constant_tags = constant_tags

        if autodiff_backend.upper() == "AUTO":
            autodiff_backend = _DEFAULT_BACKEND
        self._autodiff_backend = autodiff_backend
        self._multiloss = isinstance(loss_fn, Iterable)

        # the object that handles converting to backend + computing gradient
        if self._multiloss:
            # special meta-handler if loss function is sequence to sum
            backend_opts["executor"] = executor
            self.handler = MultiLossHandler(autodiff_backend, **backend_opts)
        else:
            self.handler = _BACKEND_HANDLERS[autodiff_backend](**backend_opts)

        # use identity if no nomalization required
        if norm_fn is None:
            norm_fn = identity_fn
        self.norm_fn = norm_fn

        self.reset(tn, loss_target=loss_target)

        # convert constant raw arrays ahead of time to correct backend
        self.loss_constants = tree_map(
            functools.partial(convert_raw_arrays, f=self.handler.to_constant),
            ensure_dict(loss_constants),
        )
        self.loss_kwargs = ensure_dict(loss_kwargs)
        kws = {**self.loss_constants, **self.loss_kwargs}

        # inject these constant options to the loss function(s)
        if self._multiloss:
            # loss is a sum of independent terms
            self.loss_fn = [functools.partial(fn, **kws) for fn in loss_fn]
        else:
            # loss is all in one
            self.loss_fn = functools.partial(loss_fn, **kws)

        # first we wrap the function to convert from array args to TN arg
        #     (i.e. to autodiff library compatible form)
        if self._multiloss:
            array_fn = [
                MakeArrayFn(self._tn_opt, fn, self.norm_fn, autodiff_backend)
                for fn in self.loss_fn
            ]
        else:
            array_fn = MakeArrayFn(
                self._tn_opt, self.loss_fn, self.norm_fn, autodiff_backend
            )

        # then we pass it to the handler which generates a function that
        # computes both the value and gradients (still in array form)
        self.handler.setup_fn(array_fn)

        # options to do with the minimizer
        self.bounds = bounds
        self.optimizer = optimizer
        self.callback = callback

    def _set_tn(self, tn):
        # work out which tensors to optimize and get the underlying data
        self._tn_opt, variables = parse_network_to_backend(
            tn,
            tags=self.tags,
            shared_tags=self.shared_tags,
            constant_tags=self.constant_tags,
            to_constant=self.handler.to_constant,
        )
        # handles storing and packing / unpacking many arrays as a vector
        self.vectorizer = Vectorizer(variables)

    def _reset_tracking_info(self, loss_target=None):
        # tracking info
        self.loss = float("inf")
        self.loss_best = float("inf")
        self.loss_target = loss_target
        self.losses = []
        self._n = 0
        self._pbar = None

    def reset(self, tn=None, clear_info=True, loss_target=None):
        """Reset this optimizer without losing the compiled loss and gradient
        functions.

        Parameters
        ----------
        tn : TensorNetwork, optional
            Set this tensor network as the current state of the optimizer, it
            must exactly match the original tensor network.
        clear_info : bool, optional
            Clear the tracked losses and iterations.
        """
        if tn is not None:
            self._set_tn(tn)
        if clear_info:
            self._reset_tracking_info(loss_target=loss_target)

    def _maybe_init_pbar(self, n):
        if self.progbar:
            self._pbar = tqdm.tqdm(total=n)

    def _maybe_update_pbar(self):
        if self._pbar is not None:
            self._pbar.update()
            self.loss_best = min(self.loss_best, self.loss)
            msg = f"{self.loss:+.12f} [best: {self.loss_best:+.12f}] "
            self._pbar.set_description(msg)

    def _maybe_close_pbar(self):
        if self._pbar is not None:
            self._pbar.close()
            self._pbar = None

    def _check_loss_target(self):
        if (self.loss_target is not None) and (self.loss <= self.loss_target):
            # for scipy terminating optimizer with callback doesn't work
            raise KeyboardInterrupt

    def _maybe_call_callback(self):
        if self.callback is not None:
            self.callback(self)

    def vectorized_value(self, x):
        """The value of the loss function at vector ``x``."""
        self.vectorizer.vector[:] = x
        arrays = self.vectorizer.unpack()
        self.loss = self.handler.value(arrays).item()
        self.losses.append(self.loss)
        self._n += 1
        self._maybe_update_pbar()
        self._check_loss_target()
        self._maybe_call_callback()
        return self.loss

    def vectorized_value_and_grad(self, x):
        """The value and gradient of the loss function at vector ``x``."""
        self.vectorizer.vector[:] = x
        arrays = self.vectorizer.unpack()
        result, grads = self.handler.value_and_grad(arrays)
        self._n += 1
        self.loss = result.item()
        self.losses.append(self.loss)
        vec_grad = self.vectorizer.pack(grads, "grad")
        self._maybe_update_pbar()
        self._check_loss_target()
        self._maybe_call_callback()
        return self.loss, vec_grad

    def vectorized_hessp(self, x, p):
        """The action of the hessian at point ``x`` on vector ``p``."""
        primals = self.vectorizer.unpack(x)
        tangents = self.vectorizer.unpack(p)
        hp_arrays = self.handler.hessp(primals, tangents)
        self._n += 1
        self.losses.append(self.loss)
        self._maybe_update_pbar()
        return self.vectorizer.pack(hp_arrays, "hp")

    def __repr__(self):
        return (
            f"<TNOptimizer(d={self.d}, " f"backend={self._autodiff_backend})>"
        )

    @property
    def d(self):
        return int(self.vectorizer.d)

    @property
    def nevals(self):
        """The number of gradient evaluations."""
        return self._n

    @property
    def optimizer(self):
        """The underlying optimizer that works with the vectorized functions."""
        return self._optimizer

    @optimizer.setter
    def optimizer(self, x):
        self._optimizer = x
        if self.optimizer in _STOC_GRAD_METHODS:
            self._method = _STOC_GRAD_METHODS[self.optimizer]()
        else:
            self._method = self.optimizer

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, x):
        if x is not None:
            self._bounds = np.array((x,) * self.vectorizer.d)
        else:
            self._bounds = None

    def get_tn_opt(self):
        """Extract the optimized tensor network, this is a three part process:

            1. inject the current optimized vector into the target tensor
               network,
            2. run it through ``norm_fn``,
            3. drop any tags used to identify variables.

        Returns
        -------
        tn_opt : TensorNetwork
        """
        arrays = tree_map(
            self.handler.to_constant,
            self.vectorizer.unpack(),
        )
        inject_(arrays, self._tn_opt)
        tn = self.norm_fn(self._tn_opt.copy())
        tn.drop_tags(t for t in tn.tags if variable_finder.match(t))

        for t in tn:
            if isinstance(t, PTensor):
                t.params = to_numpy(t.params)
            else:
                t.modify(data=to_numpy(t.data), left_inds=t.left_inds)

        return tn

    def optimize(
        self, n, tol=None, jac=True, hessp=False, optlib="scipy", **options
    ):
        """Run the optimizer for ``n`` function evaluations, using by default
        :func:`scipy.optimize.minimize` as the driver for the vectorized
        computation. Supplying the gradient and hessian vector product is
        controlled by the ``jac`` and ``hessp`` options respectively.

        Parameters
        ----------
        n : int
            Notionally the maximum number of iterations for the optimizer, note
            that depending on the optimizer being used, this may correspond to
            number of function evaluations rather than just iterations.
        tol : None or float, optional
            Tolerance for convergence, note that various more specific
            tolerances can usually be supplied to ``options``, depending on
            the optimizer being used.
        jac : bool, optional
            Whether to supply the jacobian, i.e. gradient, of the loss
            function.
        hessp : bool, optional
            Whether to supply the hessian vector product of the loss function.
        optlib : {'scipy', 'nlopt'}, optional
            Which optimization library to use.
        options
            Supplied to :func:`scipy.optimize.minimize` or whichever optimizer
            is being used.

        Returns
        -------
        tn_opt : TensorNetwork
        """
        return {"scipy": self.optimize_scipy, "nlopt": self.optimize_nlopt,}[
            optlib
        ](n=n, tol=tol, jac=jac, hessp=hessp, **options)

    def optimize_scipy(self, n, tol=None, jac=True, hessp=False, **options):
        """Scipy based optimization, see
        :meth:`~quimb.tensor.optimize.TNOptimizer.optimize` for details.
        """
        from scipy.optimize import minimize

        if jac:
            fun = self.vectorized_value_and_grad
        else:
            fun = self.vectorized_value

        try:
            self._maybe_init_pbar(n)
            self.res = minimize(
                fun=fun,
                jac=jac,
                hessp=self.vectorized_hessp if hessp else None,
                x0=self.vectorizer.vector,
                tol=tol,
                bounds=self.bounds,
                method=self._method,
                options=dict(maxiter=n, **options),
            )
            self.vectorizer.vector[:] = self.res.x
        except KeyboardInterrupt:
            pass
        finally:
            self._maybe_close_pbar()

        return self.get_tn_opt()

    def optimize_basinhopping(
        self, n, nhop, temperature=1.0, jac=True, hessp=False, **options
    ):
        """Run the optimizer for using :func:`scipy.optimize.basinhopping`
        as the driver for the vectorized computation. This performs ``nhop``
        local optimization each with ``n`` iterations.

        Parameters
        ----------
        n : int
            Number of iterations per local optimization.
        nhop : int
            Number of local optimizations to hop between.
        temperature : float, optional
            H
        options
            Supplied to the inner :func:`scipy.optimize.minimize` call.

        Returns
        -------
        tn_opt : TensorNetwork
        """
        from scipy.optimize import basinhopping

        if jac:
            fun = self.vectorized_value_and_grad
        else:
            fun = self.vectorized_value

        try:
            self._maybe_init_pbar(n * nhop)
            self.res = basinhopping(
                func=fun,
                x0=self.vectorizer.vector,
                niter=nhop,
                minimizer_kwargs=dict(
                    jac=jac,
                    hessp=self.vectorized_hessp if hessp else None,
                    method=self._method,
                    bounds=self.bounds,
                    options=dict(maxiter=n, **options),
                ),
                T=temperature,
            )
            self.vectorizer.vector[:] = self.res.x

        except KeyboardInterrupt:
            pass
        finally:
            self._maybe_close_pbar()

        return self.get_tn_opt()

    def optimize_nlopt(
        self,
        n,
        tol=None,
        jac=True,
        hessp=False,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        xtol_abs=None,
    ):
        """Run the optimizer for ``n`` function evaluations, using ``nlopt`` as
        the backend library to run the optimization. Whether the gradient is
        computed depends on which ``optimizer`` is selected, see valid options
        at  https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/.

        Parameters
        ----------
        n : int
            The maximum number of iterations for the optimizer.
        tol : None or float, optional
            Tolerance for convergence, here this is taken to be the relative
            tolerance for the loss (``ftol_rel`` below overrides this).
        jac : bool, optional
            Whether to supply the jacobian, i.e. gradient, of the loss
            function.
        hessp : bool, optional
            Whether to supply the hessian vector product of the loss function.
        ftol_rel : float, optional
            Set relative tolerance on function value.
        ftol_abs : float, optional
            Set absolute tolerance on function value.
        xtol_rel : float, optional
            Set relative tolerance on optimization parameters.
        xtol_abs : float, optional
            Set absolute tolerances on optimization parameters.

        Returns
        -------
        tn_opt : TensorNetwork
        """
        import nlopt

        if not (jac is True and hessp is False):
            raise NotImplementedError(
                "Only gradient based optimizers are "
                "supported with nlopt currently."
            )

        if tol == 0.0:
            # assume no stopping criteria
            ftol_rel == 0.0 if ftol_rel is None else ftol_rel
            ftol_abs == 0.0 if ftol_abs is None else ftol_abs
            xtol_rel == 0.0 if xtol_rel is None else xtol_rel
            xtol_abs == 0.0 if xtol_abs is None else xtol_abs

        elif (tol is not None) and (ftol_rel is None):
            # assume relative loss tolerance is specified
            ftol_rel = tol

        # translate directly comparable algorithms
        optimizer = {
            "L-BFGS-B": "LD_LBFGS",
            "COBYLA": "LN_COBYLA",
            "SLSQP": "LD_SLSQP",
        }.get(self.optimizer.upper(), self.optimizer)

        try:
            self._maybe_init_pbar(n)

            def f(x, grad):
                self.vectorizer.vector[:] = x
                arrays = self.vectorizer.unpack()
                if grad.size > 0:
                    result, grads = self.handler.value_and_grad(arrays)
                    grad[:] = self.vectorizer.pack(grads, "grad")
                else:
                    result = self.handler.value(arrays)
                self._n += 1
                self.loss = result.item()
                self.losses.append(self.loss)
                self._maybe_update_pbar()
                return self.loss

            opt = nlopt.opt(getattr(nlopt, optimizer), self.d)
            opt.set_min_objective(f)
            opt.set_maxeval(n)

            if self.bounds is not None:
                opt.set_lower_bounds(self.bounds[:, 0])
                opt.set_upper_bounds(self.bounds[:, 1])

            if self.loss_target is not None:
                opt.set_stopval(self.loss_target)
            if ftol_rel is not None:
                opt.set_ftol_rel(ftol_rel)
            if ftol_abs is not None:
                opt.set_ftol_abs(ftol_abs)
            if xtol_rel is not None:
                opt.set_xtol_rel(xtol_rel)
            if xtol_abs is not None:
                opt.set_xtol_abs(xtol_abs)

            self.vectorizer.vector[:] = opt.optimize(self.vectorizer.vector)

        except (KeyboardInterrupt, RuntimeError):
            pass
        finally:
            self._maybe_close_pbar()

        return self.get_tn_opt()

    def optimize_ipopt(self, n, tol=None, **options):
        """Run the optimizer for ``n`` function evaluations, using ``ipopt`` as
        the backend library to run the optimization via the python package
        ``cyipopt``.

        Parameters
        ----------
        n : int
            The maximum number of iterations for the optimizer.

        Returns
        -------
        tn_opt : TensorNetwork
        """
        from cyipopt import minimize_ipopt

        try:
            self._maybe_init_pbar(n)
            self.res = minimize_ipopt(
                fun=self.vectorized_value_and_grad,
                jac=True,
                x0=self.vectorizer.vector,
                tol=tol,
                bounds=self.bounds,
                method=self._method,
                options=dict(maxiter=n, **options),
            )
            self.vectorizer.vector[:] = self.res.x
        except KeyboardInterrupt:
            pass
        finally:
            self._maybe_close_pbar()

        return self.get_tn_opt()

    def optimize_nevergrad(self, n):
        """Run the optimizer for ``n`` function evaluations, using
        ``nevergrad`` as the backend library to run the optimization. As the
        name suggests, the gradient is not required for this method.

        Parameters
        ----------
        n : int
            The maximum number of iterations for the optimizer.

        Returns
        -------
        tn_opt : TensorNetwork
        """
        import nevergrad as ng

        opt = getattr(ng.optimizers, self.optimizer)(
            parametrization=ng.p.Array(
                init=self.vectorizer.vector,
                lower=self.bounds[:, 0] if self.bounds is not None else None,
                upper=self.bounds[:, 1] if self.bounds is not None else None,
            ),
            budget=n,
        )

        try:
            self._maybe_init_pbar(n)
            for _ in range(n):
                x = opt.ask()
                loss = self.vectorized_value(*x.args, **x.kwargs)
                opt.tell(x, loss)
                if self.loss_target is not None:
                    if self.loss < self.loss_target:
                        break

        except KeyboardInterrupt:
            pass
        finally:
            self._maybe_close_pbar()

        # recommendation = opt.minimize(self.vectorized_value)
        recommendation = opt.provide_recommendation()
        self.vectorizer.vector[:] = recommendation.value

        return self.get_tn_opt()



import pytest
import itertools
import math
import numpy as np
from numpy.testing import assert_allclose
import quimb as qu


@pytest.fixture
def p1():
    return qu.rand_rho(3)


@pytest.fixture
def p2():
    return qu.rand_rho(3)


@pytest.fixture
def k1():
    return qu.rand_ket(3)


@pytest.fixture
def k2():
    return qu.rand_ket(3)


@pytest.fixture
def orthog_ks():
    p = qu.rand_rho(3)
    v = qu.eigvecsh(p)
    return (v[:, [0]], v[:, [1]], v[:, [2]])


# --------------------------------------------------------------------------- #
# TESTS                                                                       #
# --------------------------------------------------------------------------- #

class TestFidelity:
    def test_both_pure(self, k1, k2):
        f = qu.fidelity(k1, k1)
        assert_allclose(f, 1.0)
        f = qu.fidelity(k1, k2)
        assert f > 0 and f < 1

    def test_both_mixed(self, p1, p2):
        f = qu.fidelity(qu.eye(3) / 3, qu.eye(3) / 3)
        assert_allclose(f, 1.0)
        f = qu.fidelity(p1, p1)
        assert_allclose(f, 1.0)
        f = qu.fidelity(p1, p2)
        assert f > 0 and f < 1

    def test_orthog_pure(self, orthog_ks):
        k1, k2, k3 = orthog_ks
        for s1, s2, in ([k1, k2],
                        [k2, k3],
                        [k3, k1],
                        [k1 @ k1.H, k2],
                        [k1, k2 @ k2.H],
                        [k3 @ k3.H, k2],
                        [k3, k2 @ k2.H],
                        [k1 @ k1.H, k3],
                        [k1, k3 @ k3.H],
                        [k1 @ k1.H, k2 @ k2.H],
                        [k2 @ k2.H, k3 @ k3.H],
                        [k1 @ k1.H, k3 @ k3.H]):
            f = qu.fidelity(s1, s2)
            assert_allclose(f, 0.0, atol=1e-6)


class TestPurify:
    def test_d2(self):
        rho = qu.eye(2) / 2
        psi = qu.purify(rho)
        assert qu.expec(psi, qu.bell_state('phi+')) > 1 - 1e-14

    def test_pure(self):
        rho = qu.up(qtype='dop')
        psi = qu.purify(rho)
        assert abs(qu.concurrence(psi)) < 1e-14


class TestDephase:
    @pytest.mark.parametrize("rand_rank", [None, 0.3, 2])
    def test_basic(self, rand_rank):
        rho = qu.rand_rho(9)
        ln = qu.logneg(rho, [3, 3])
        for p in (0.2, 0.5, 0.8, 1.0):
            rho_d = qu.dephase(rho, p, rand_rank=rand_rank)
            assert qu.logneg(rho_d, [3, 3]) <= ln
            assert rho_d.tr() == pytest.approx(1.0)


class TestKrausOp:

    @pytest.mark.parametrize("stack", [False, True])
    def test_depolarize(self, stack):
        rho = qu.rand_rho(2)
        I, X, Y, Z = (qu.pauli(s) for s in 'IXYZ')
        es = [qu.expec(rho, A) for A in (X, Y, Z)]
        p = 0.1
        Ek = [(1 - p)**0.5 * I,
              (p / 3)**0.5 * X,
              (p / 3)**0.5 * Y,
              (p / 3)**0.5 * Z]
        if stack:
            Ek = np.stack(Ek, axis=0)
        sigma = qu.kraus_op(rho, Ek, check=True)
        es2 = [qu.expec(sigma, A) for A in (X, Y, Z)]
        assert qu.tr(sigma) == pytest.approx(1.0)
        assert all(abs(e2) < abs(e) for e, e2 in zip(es, es2))
        sig_exp = sum(E @ rho @ qu.dag(E) for E in Ek)
        assert_allclose(sig_exp, sigma)

    def test_subsystem(self):
        rho = qu.rand_rho(6)
        dims = [3, 2]
        I, X, Y, Z = (qu.pauli(s) for s in 'IXYZ')
        mi_i = qu.mutual_information(rho, dims)
        p = 0.1
        Ek = [(1 - p)**0.5 * I,
              (p / 3)**0.5 * X,
              (p / 3)**0.5 * Y,
              (p / 3)**0.5 * Z]

        with pytest.raises(ValueError):
            qu.kraus_op(rho, qu.randn((3, 2, 2)), check=True,
                        dims=dims, where=1)

        sigma = qu.kraus_op(rho, Ek, check=True, dims=dims, where=1)
        mi_f = qu.mutual_information(sigma, dims)
        assert mi_f < mi_i
        assert qu.tr(sigma) == pytest.approx(1.0)
        sig_exp = sum((qu.eye(3) & E) @ rho @ qu.dag(qu.eye(3) & E)
                      for E in Ek)
        assert_allclose(sig_exp, sigma)

    def test_multisubsystem(self):
        qu.seed_rand(42)
        dims = [2, 2, 2]
        IIX = qu.ikron(qu.rand_matrix(2), dims, 2)
        dcmp = qu.pauli_decomp(IIX, mode='c')
        for p, x in dcmp.items():
            if abs(x) < 1e-12:
                assert (p[0] != 'I') or (p[1] != 'I')
            else:
                assert p[0] == p[1] == 'I'
        K = qu.rand_iso(3 * 4, 4).reshape(3, 4, 4)
        KIIXK = qu.kraus_op(IIX, K, dims=dims, where=[0, 2])
        dcmp = qu.pauli_decomp(KIIXK, mode='c')
        for p, x in dcmp.items():
             if abs(x) > 1e-12:
                assert (p == 'III') or p[0] != 'I'


class TestProjector:

    def test_simple(self):
        Z = qu.pauli('Z')
        P = qu.projector(Z & Z)
        uu = qu.dop(qu.up()) & qu.dop(qu.up())
        dd = qu.dop(qu.down()) & qu.dop(qu.down())
        assert_allclose(P, uu + dd)
        assert qu.expec(P, qu.bell_state('phi+')) == pytest.approx(1.0)
        assert qu.expec(P, qu.bell_state('psi+')) == pytest.approx(0.0)


class TestMeasure:

    def test_pure(self):
        psi = qu.bell_state('psi-')
        IZ = qu.pauli('I') & qu.pauli('Z')
        ZI = qu.pauli('Z') & qu.pauli('I')
        res, psi_after = qu.measure(psi, IZ)
        # normalized
        assert qu.expectation(psi_after, psi_after) == pytest.approx(1.0)
        # anticorrelated
        assert qu.expectation(psi_after, IZ) == pytest.approx(res)
        assert qu.expectation(psi_after, ZI) == pytest.approx(-res)
        assert isinstance(psi_after, qu.qarray)

    def test_bigger(self):
        psi = qu.rand_ket(2**5)
        assert np.sum(abs(psi) < 1e-12) == 0
        A = qu.kronpow(qu.pauli('Z'), 5)
        res, psi_after = qu.measure(psi, A, eigenvalue=-1.0)
        # should have projected to half subspace
        assert np.sum(abs(psi_after) < 1e-12) == 2**4
        assert res == -1.0

    def test_mixed(self):
        rho = qu.dop(qu.bell_state('psi-'))
        IZ = qu.pauli('I') & qu.pauli('Z')
        ZI = qu.pauli('Z') & qu.pauli('I')
        res, rho_after = qu.measure(rho, IZ)
        # normalized
        assert qu.tr(rho_after) == pytest.approx(1.0)
        # anticorrelated
        assert qu.expectation(rho_after, IZ) == pytest.approx(res)
        assert qu.expectation(rho_after, ZI) == pytest.approx(-res)
        assert isinstance(rho_after, qu.qarray)


class TestSimulateCounts:

    @pytest.mark.parametrize('qtype', ['ket', 'dop'])
    def test_ghz(self, qtype):
        psi = qu.ghz_state(3, qtype=qtype)
        results = qu.simulate_counts(psi, 1024)
        assert len(results) == 2
        assert '000' in results
        assert '111' in results


class TestCPrint:

    def test_basic(self):
        psi = qu.ghz_state(2)
        qu.cprint(psi)


class TestEntropy:
    def test_entropy_pure(self):
        a = qu.bell_state(1, qtype='dop')
        assert_allclose(0.0, qu.entropy(a), atol=1e-12)

    def test_entropy_mixed(self):
        a = 0.5 * (qu.bell_state(1, qtype='dop') +
                   qu.bell_state(2, qtype='dop'))
        assert_allclose(1.0, qu.entropy(a), atol=1e-12)

    @pytest.mark.parametrize("evals, e", [([0, 1, 0, 0], 0),
                                          ([0, 0.5, 0, 0.5], 1),
                                          ([0.25, 0.25, 0.25, 0.25], 2)])
    def test_list(self, evals, e):
        assert_allclose(qu.entropy(evals), e)

    @pytest.mark.parametrize("evals, e", [([0, 1, 0, 0], 0),
                                          ([0, 0.5, 0, 0.5], 1),
                                          ([0.25, 0.25, 0.25, 0.25], 2)])
    def test_1darray(self, evals, e):
        assert_allclose(qu.entropy(np.asarray(evals)), e)

    @pytest.mark.parametrize("m", [1, 2, 3])
    def test_rank(self, m):
        k = qu.rand_ket(2**4)
        pab = qu.ptr(k, [2, 2, 2, 2], range(m))
        ef = qu.entropy(pab)
        er = qu.entropy(pab, rank=2**m)
        assert_allclose(ef, er)

    def test_entropy_subsystem(self):
        p = qu.rand_ket(2**9)
        # exact
        e1 = qu.entropy_subsys(p, (2**5, 2**4), 0, approx_thresh=1e30)
        # approx
        e2 = qu.entropy_subsys(p, (2**5, 2**4), 0, approx_thresh=1)
        assert e1 != e2
        assert_allclose(e1, e2, rtol=0.2)

        assert qu.entropy_subsys(p, (2**5, 2**4), [0, 1],
                                 approx_thresh=1) == 0.0


class TestMutualInformation:
    def test_mutual_information_pure(self):
        a = qu.bell_state(0)
        assert_allclose(qu.mutual_information(a), 2.)
        a = qu.rand_product_state(2)
        assert_allclose(qu.mutual_information(a), 0., atol=1e-12)

    def test_mutual_information_pure_sub(self):
        a = qu.up() & qu.bell_state(1)
        ixy = qu.mutual_information(a, [2, 2, 2], 0, 1)
        assert_allclose(0.0, ixy, atol=1e-12)
        ixy = qu.mutual_information(a, [2, 2, 2], 0, 2)
        assert_allclose(0.0, ixy, atol=1e-12)
        ixy = qu.mutual_information(a, [2, 2, 2], 2, 1)
        assert_allclose(2.0, ixy, atol=1e-12)

    @pytest.mark.parametrize('inds', [(0, 1), (1, 2), (0, 2)])
    def test_mixed_sub(self, inds):
        a = qu.rand_rho(2**3)
        rho_ab = qu.ptr(a, [2, 2, 2], inds)
        ixy = qu.mutual_information(rho_ab, (2, 2))
        assert (0 <= ixy <= 2.0)

    def test_mutinf_interleave(self):
        p = qu.dop(qu.singlet() & qu.singlet())
        ixy = qu.mutual_information(p, [2] * 4, sysa=(0, 2))
        assert_allclose(ixy, 4)

    def test_mutinf_interleave_pure(self):
        p = qu.singlet() & qu.singlet()
        ixy = qu.mutual_information(p, [2] * 4, sysa=(0, 2))
        assert_allclose(ixy, 4)

    def test_mutinf_subsys(self):
        p = qu.rand_ket(2**9)
        dims = (2**3, 2**2, 2**4)
        # exact
        rho_ab = qu.ptr(p, dims, [0, 2])
        mi0 = qu.mutual_information(rho_ab, [8, 16])
        mi1 = qu.mutinf_subsys(p, dims, sysa=0, sysb=2, approx_thresh=1e30)
        assert_allclose(mi1, mi0)
        # approx
        mi2 = qu.mutinf_subsys(p, dims, sysa=0, sysb=2, approx_thresh=1)
        assert_allclose(mi1, mi2, rtol=0.1)

    def test_mutinf_subsys_pure(self):
        p = qu.rand_ket(2**7)
        dims = (2**3, 2**4)
        # exact
        mi0 = qu.mutual_information(p, dims, sysa=0)
        mi1 = qu.mutinf_subsys(p, dims, sysa=0, sysb=1, approx_thresh=1e30)
        assert_allclose(mi1, mi0)
        # approx
        mi2 = qu.mutinf_subsys(p, dims, sysa=0, sysb=1,
                               approx_thresh=1, tol=5e-3)
        assert_allclose(mi1, mi2, rtol=0.1)


class TestSchmidtGap:
    def test_bell_state(self):
        p = qu.bell_state('psi-')
        assert_allclose(qu.schmidt_gap(p, [2, 2], 0), 0.0)
        p = qu.up() & qu.down()
        assert_allclose(qu.schmidt_gap(p, [2, 2], 0), 1.0)
        p = qu.rand_ket(2**3)
        assert 0 < qu.schmidt_gap(p, [2] * 3, sysa=[0, 1]) < 1.0


class TestPartialTranspose:
    def test_partial_transpose(self):
        a = qu.bell_state(0, qtype='dop')
        b = qu.partial_transpose(a)
        assert isinstance(b, qu.qarray)
        assert_allclose(b, np.array([[0, 0, 0, -0.5],
                                     [0, 0.5, 0, 0],
                                     [0, 0, 0.5, 0],
                                     [-0.5, 0, 0, 0]]))

    def test_tr_sqrt_rank(self):
        psi = qu.rand_ket(2**5)
        rhoa = psi.ptr([2] * 5, range(4))
        assert_allclose(qu.tr_sqrt(rhoa), qu.tr_sqrt(rhoa, rank=2))


class TestNegativity:
    @pytest.mark.parametrize("bs", ['psi-', 'phi-', 'psi+', 'phi+'])
    @pytest.mark.parametrize("qtype", ['ket', 'dop'])
    def test_simple(self, qtype, bs):
        p = qu.bell_state(bs, qtype=qtype)
        assert qu.negativity(p) > 0.5 - 1e-14

    def test_subsystem(self):
        p = qu.singlet_pairs(4)
        rhoab = p.ptr([2, 2, 2, 2], [0, 1])
        assert qu.negativity(rhoab, [2] * 2) > 0.5 - 1e-14
        rhoab = p.ptr([2, 2, 2, 2], [1, 2])
        assert qu.negativity(rhoab, [2] * 2) < 1e-14
        rhoab = p.ptr([2, 2, 2, 2], [2, 3])
        assert qu.negativity(rhoab, [2] * 2) > 0.5 - 1e-14


class TestLogarithmicNegativity:
    @pytest.mark.parametrize("bs", ['psi-', 'phi-', 'psi+', 'phi+'])
    @pytest.mark.parametrize("qtype", ['ket', 'dop'])
    def test_bell_states(self, qtype, bs):
        p = qu.bell_state(bs, qtype=qtype)
        assert qu.logneg(p) > 1.0 - 1e-14

    def test_subsystem(self):
        p = qu.singlet_pairs(4)
        rhoab = p.ptr([2, 2, 2, 2], [0, 1])
        assert qu.logneg(rhoab, [2] * 2) > 1 - 1e-14
        rhoab = p.ptr([2, 2, 2, 2], [1, 2])
        assert qu.logneg(rhoab, [2] * 2) < 1e-14
        rhoab = p.ptr([2, 2, 2, 2], [2, 3])
        assert qu.logneg(rhoab, [2] * 2) > 1 - 1e-14

    def test_interleaving(self):
        p = qu.permute(qu.singlet() & qu.singlet(),
                       [2, 2, 2, 2], [0, 2, 1, 3])
        assert qu.logneg(p, [2] * 4, sysa=[0, 3]) > 2 - 1e-13

    def test_logneg_subsys(self):
        p = qu.rand_ket(2**(2 + 3 + 1 + 2))
        dims = (2**2, 2**3, 2**1, 2**2)
        sysa = [0, 3]
        sysb = 1
        # exact 1
        ln0 = qu.logneg(qu.ptr(p, dims, [0, 1, 3]), [4, 8, 4], [0, 2])
        # exact 2
        ln1 = qu.logneg_subsys(p, dims, sysa, sysb, approx_thresh=1e30)
        assert_allclose(ln0, ln1)
        # approx
        ln2 = qu.logneg_subsys(p, dims, sysa, sysb, approx_thresh=1)
        assert ln1 != ln2
        assert_allclose(ln1, ln2, rtol=5e-2)

    def test_logneg_subsys_pure(self):
        p = qu.rand_ket(2**(3 + 4))
        dims = (2**3, 2**4)
        sysa = 0
        sysb = 1
        # exact 1
        ln0 = qu.logneg(p, dims, 0)
        # exact 2
        ln1 = qu.logneg_subsys(p, dims, sysa, sysb, approx_thresh=1e30)
        assert_allclose(ln0, ln1)
        # approx
        ln2 = qu.logneg_subsys(p, dims, sysa, sysb, approx_thresh=1, tol=5e-3)
        assert ln1 != ln2
        assert_allclose(ln1, ln2, rtol=1e-1)

    def test_logneg_subsys_pure_should_swap_subsys(self):
        p = qu.rand_ket(2**(5 + 2))
        dims = (2**5, 2**2)
        sysa = 0
        sysb = 1
        # exact 1
        ln0 = qu.logneg(p, dims, 0)
        # exact 2
        ln1 = qu.logneg_subsys(p, dims, sysa, sysb, approx_thresh=1e30)
        assert_allclose(ln0, ln1)
        # approx
        ln2 = qu.logneg_subsys(p, dims, sysa, sysb, approx_thresh=1, tol=0.005)
        assert ln1 != ln2
        assert_allclose(ln1, ln2, rtol=0.2)


class TestConcurrence:
    @pytest.mark.parametrize("bs", ['psi-', 'phi-', 'psi+', 'phi+'])
    @pytest.mark.parametrize("qtype", ['ket', 'dop'])
    def test_bell_states(self, qtype, bs):
        p = qu.bell_state(bs, qtype=qtype)
        assert qu.concurrence(p) > 1.0 - 1e-14

    def test_subsystem(self):
        p = qu.rand_rho(2**4)
        e = qu.concurrence(p, [2, 2, 2, 2], 1, 2)
        assert 0 <= e <= 1


class TestQuantumDiscord:
    def test_owci(self):
        a = qu.qu([1, 0], qtype='op')
        b = qu.qu([0, 1], qtype='op')
        for _ in (0, 1, 2, 3):
            p = qu.rand_product_state(2)
            ci = qu.one_way_classical_information(p @ p.H, [a, b])
            assert_allclose(ci, 0., atol=1e-12)
        for i in (0, 1, 2, 3):
            p = qu.bell_state(i)
            ci = qu.one_way_classical_information(p @ p.H, [a, b])
            assert_allclose(ci, 1., atol=1e-12)

    def test_quantum_discord_sep(self):
        for _ in range(10):
            p = qu.rand_product_state(2)
            p = p @ p.H
            qd = qu.quantum_discord(p)
            assert_allclose(0.0, qd, atol=1e-12)

    def test_quantum_discord_pure(self):
        for _ in range(10):
            p = qu.rand_ket(4)
            p = p @ p.H
            iab = qu.mutual_information(p)
            qd = qu.quantum_discord(p)
            assert_allclose(iab / 2, qd)

    def test_quantum_discord_mixed(self):
        for _ in range(10):
            p = qu.rand_mix(4)
            p = p @ p.H
            qd = qu.quantum_discord(p)
            assert(0 <= qd and qd <= 1)

    def test_auto_trace_out(self):
        p = qu.rand_rho(2**3)
        qd = qu.quantum_discord(p, [2, 2, 2], 0, 2)
        assert(0 <= qd and qd <= 1)


class TestTraceDistance:
    def test_types(self, k1, k2):
        td1 = qu.trace_distance(k1, k2)
        td2 = qu.trace_distance(qu.dop(k1), k2)
        td3 = qu.trace_distance(k1, qu.dop(k2))
        td4 = qu.trace_distance(qu.dop(k1), qu.dop(k2))
        assert_allclose([td1] * 3, [td2, td3, td4])

    def test_same(self, p1):
        assert abs(qu.trace_distance(p1, p1)) < 1e-14

    @pytest.mark.parametrize("uqtype", ['ket', 'dop'])
    @pytest.mark.parametrize("dqtype", ['ket', 'dop'])
    def test_distinguishable(self, uqtype, dqtype):
        assert qu.trace_distance(qu.up(qtype=uqtype),
                                 qu.down(qtype=dqtype)) > 1 - 1e-10


class TestDecomp:
    @pytest.mark.parametrize("qtype", ['ket', 'dop'])
    def test_pauli_decomp_singlet(self, qtype):
        p = qu.singlet(qtype=qtype)
        names_cffs = qu.pauli_decomp(p, mode='cp')
        assert_allclose(names_cffs['II'], 0.25)
        assert_allclose(names_cffs['ZZ'], -0.25)
        assert_allclose(names_cffs['YY'], -0.25)
        assert_allclose(names_cffs['ZZ'], -0.25)
        for name in itertools.permutations('IXYZ', 2):
            assert_allclose(names_cffs["".join(name)], 0.0)

    def test_pauli_reconstruct(self):
        p1 = qu.rand_rho(4)
        names_cffs = qu.pauli_decomp(p1, mode='c')
        pr = sum(
            qu.kron(*(qu.pauli(s) for s in name)) * names_cffs["".join(name)]
            for name in itertools.product('IXYZ', repeat=2)
        )
        assert_allclose(pr, p1)

    @pytest.mark.parametrize(
        "state, out",
        [(qu.up() & qu.down(), {0: 0.5, 1: 0.5, 2: 0, 3: 0}),
         (qu.down() & qu.down(), {0: 0, 1: 0, 2: 0.5, 3: 0.5}),
         (qu.singlet() & qu.singlet(), {'00': 1.0, '23': 0.0})])
    def test_bell_decomp(self, state, out):
        names_cffs = qu.bell_decomp(state, mode='c')
        for key in out:
            assert_allclose(names_cffs[str(key)], out[key])


class TestCorrelation:
    @pytest.mark.parametrize("pre_c", [False, True])
    @pytest.mark.parametrize("p_sps", [True, False])
    @pytest.mark.parametrize("op_sps", [True, False])
    @pytest.mark.parametrize("dims", (None, [2, 2]))
    def test_types(self, dims, op_sps, p_sps, pre_c):
        p = qu.rand_rho(4, sparse=p_sps)
        c = qu.correlation(p, qu.pauli('x', sparse=op_sps),
                           qu.pauli('z', sparse=op_sps),
                           0, 1, dims=dims, precomp_func=pre_c)
        c = c(p) if pre_c else c
        assert c >= -1.0
        assert c <= 1.0

    @pytest.mark.parametrize("pre_c", [False, True])
    @pytest.mark.parametrize("qtype", ["ket", "dop"])
    @pytest.mark.parametrize("s", ['x', 'y', 'z'])
    def test_classically_no_correlated(self, s, qtype, pre_c):
        p = qu.up(qtype=qtype) & qu.up(qtype=qtype)
        c = qu.correlation(p, qu.pauli(s), qu.pauli(s),
                           0, 1, precomp_func=pre_c)
        c = c(p) if pre_c else c
        assert_allclose(c, 0.0)

    @pytest.mark.parametrize("pre_c", [False, True])
    @pytest.mark.parametrize("s, ct", [('x', 0), ('y', 0), ('z', 1)])
    def test_classically_correlated(self, s, ct, pre_c):
        p = 0.5 * ((qu.up(qtype='dop') & qu.up(qtype='dop')) +
                   (qu.down(qtype='dop') & qu.down(qtype='dop')))
        c = qu.correlation(p, qu.pauli(s), qu.pauli(s),
                           0, 1, precomp_func=pre_c)
        c = c(p) if pre_c else c
        assert_allclose(c, ct)

    @pytest.mark.parametrize("pre_c", [False, True])
    @pytest.mark.parametrize("s, ct", [('x', -1), ('y', -1), ('z', -1)])
    def test_entangled(self, s, ct, pre_c):
        p = qu.bell_state('psi-')
        c = qu.correlation(p, qu.pauli(s), qu.pauli(s),
                           0, 1, precomp_func=pre_c)
        c = c(p) if pre_c else c
        assert_allclose(c, ct)

    def test_reuse_precomp(self):
        cfn = qu.correlation(None, qu.pauli('z'), qu.pauli('z'), 0, 1,
                             dims=[2, 2], precomp_func=True)
        assert_allclose(cfn(qu.bell_state('psi-')), -1.0)
        assert_allclose(cfn(qu.bell_state('phi+')), 1.0)

    @pytest.mark.parametrize("pre_c", [False, True])
    def test_pauli_correlations_sum_abs(self, pre_c):
        p = qu.bell_state('psi-')
        ct = qu.pauli_correlations(p, sum_abs=True, precomp_func=pre_c)
        ct = ct(p) if pre_c else ct
        assert_allclose(ct, 3.0)

    @pytest.mark.parametrize("pre_c", [False, True])
    def test_pauli_correlations_no_sum_abs(self, pre_c):
        p = qu.bell_state('psi-')
        ct = qu.pauli_correlations(p, sum_abs=False, precomp_func=pre_c)
        assert_allclose(list(c(p) for c in ct) if pre_c else ct, (-1, -1, -1))


class TestEntCrossMatrix:
    def test_bell_state(self):
        p = qu.bell_state('phi+')
        ecm = qu.ent_cross_matrix(p, ent_fn=qu.concurrence, calc_self_ent=True)
        assert_allclose(ecm, [[1, 1], [1, 1]])

    def test_bell_state_no_self_ent(self):
        p = qu.bell_state('phi+')
        ecm = qu.ent_cross_matrix(p, ent_fn=qu.concurrence,
                                  calc_self_ent=False)
        assert_allclose(ecm, [[np.nan, 1], [1, np.nan]])

    def test_block2(self):
        p = qu.bell_state('phi+') & qu.bell_state('phi+')
        ecm = qu.ent_cross_matrix(p, ent_fn=qu.logneg, sz_blc=2)
        assert_allclose(ecm[1, 1], 0)
        assert_allclose(ecm[0, 1], 0)
        assert_allclose(ecm[1, 0], 0)

    def test_block2_no_self_ent(self):
        p = qu.bell_state('phi+') & qu.bell_state('phi+')
        ecm = qu.ent_cross_matrix(p, ent_fn=qu.logneg,
                                  calc_self_ent=False, sz_blc=2)
        assert_allclose(ecm[0, 1], 0)
        assert_allclose(ecm[0, 0], np.nan)
        assert_allclose(ecm[1, 0], 0)

    def test_block2_upscale(self):
        p = qu.bell_state('phi+') & qu.bell_state('phi+')
        ecm = qu.ent_cross_matrix(p, ent_fn=qu.logneg,
                                  calc_self_ent=False, sz_blc=2)
        assert ecm.shape == (2, 2)
        ecm = qu.ent_cross_matrix(p, ent_fn=qu.logneg, calc_self_ent=False,
                                  sz_blc=2, upscale=True)
        assert ecm.shape == (4, 4)


class TestEntCrossMatrixBlocked:
    @pytest.mark.parametrize("sz_p", [2**2 for i in [2, 3, 4, 5, 6, 9, 12]])
    @pytest.mark.parametrize("sz_blc", [1, 2, 3, 4, 5])
    @pytest.mark.parametrize("calc_self_ent", [True, False])
    def test_shapes_and_blocks(self, sz_blc, sz_p, calc_self_ent):
        if sz_p // sz_blc > 0:
            p = qu.rand_rho(2**sz_p)
            n = sz_p // sz_blc
            ecm = qu.ent_cross_matrix(p, sz_blc, calc_self_ent=calc_self_ent)
            assert ecm.shape[0] == n
            if not calc_self_ent:
                assert_allclose(np.diag(ecm), [np.nan] * n, equal_nan=True)


class TestQID:
    @pytest.mark.parametrize("bs", [0, 1, 2, 3])
    @pytest.mark.parametrize("pre_c", [False, True])
    def test_bell_state(self, bs, pre_c):
        p = qu.bell_state(bs)
        qids = qu.qid(p, dims=[2, 2], inds=[0, 1], precomp_func=pre_c)
        assert_allclose(qids(p) if pre_c else qids, [3, 3])

    @pytest.mark.parametrize("pre_c", [False, True])
    def test_random_product_state(self, pre_c):
        p = qu.rand_product_state(3)
        qids = qu.qid(p, dims=[2, 2, 2], inds=[0, 1, 2], precomp_func=pre_c)
        assert_allclose(qids(p) if pre_c else qids, [2, 2, 2])


class TestIsDegenerate:
    def test_known_degenerate(self):
        h = qu.ham_heis(2)
        assert qu.is_degenerate(h) == 2

    def test_known_nondegen(self):
        h = qu.ham_heis(2, b=0.3)
        assert qu.is_degenerate(h) == 0

    def test_supply_list(self):
        evals = [0, 1, 2, 2.0, 3]
        assert qu.is_degenerate(evals)

    def test_tol(self):
        evals = [0, 1, 1.001, 3, 4, 5, 6, 7, 8, 9]
        assert not qu.is_degenerate(evals)
        assert qu.is_degenerate(evals, tol=1e-2)


class TestPageEntropy:
    def test_known_qubit_qubit(self):
        assert abs(qu.page_entropy(2, 4) - 0.4808983469629878) < 1e-12

    def test_large_m_approx(self):
        pe = qu.page_entropy(2**10, 2**20)
        ae = 0.5 * (20 - math.log2(math.e))

        assert abs(pe - ae) < 1e-5

    def test_bigger_than_half(self):
        assert_allclose(qu.page_entropy(4, 24), qu.page_entropy(6, 24))


class TestIsEigenvector:

    def test_dense_true(self):
        a = qu.rand_herm(10)
        v = qu.eigvecsh(a)
        for i in range(10):
            assert qu.is_eigenvector(v[:, [i]], a)

    def test_dense_false(self):
        a = qu.rand_herm(10)
        v = qu.rand_ket(10)
        assert not qu.is_eigenvector(v, a)

    def test_sparse(self):
        a = qu.rand_herm(10, sparse=True, density=0.9)
        vt = qu.eigvecsh(a, sigma=0, k=1)
        assert qu.is_eigenvector(vt, a)
        vf = qu.rand_ket(10)
        assert not qu.is_eigenvector(vf, a)



import math
import itertools

import pytest
import numpy as np
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


def rand_reg_graph(reg, n, seed=None):
    import networkx as nx
    G = nx.random_regular_graph(reg, n, seed=seed)
    return G


def graph_to_qasm(G, gamma0=-0.743043, beta0=0.754082):
    n = G.number_of_nodes()

    # add all the gates
    circ = f"{n}\n"
    for i in range(n):
        circ += f"H {i}\n"
    for i, j in G.edges:
        circ += f"Rzz {gamma0} {i} {j}\n"
    for i in range(n):
        circ += f"Rx {beta0} {i}\n"

    return circ


def random_a2a_circ(L, depth, seed=42):
    rng = np.random.default_rng(seed)

    qubits = np.arange(L)
    gates = []

    for i in range(L):
        gates.append((0, 'h', i))

    for d in range(depth):
        rng.shuffle(qubits)

        for i in range(0, L - 1, 2):
            g = rng.choice(['cx', 'cy', 'cz', 'iswap'])
            gates.append((d, g, qubits[i], qubits[i + 1]))

        for q in qubits:
            g = rng.choice(['rx', 'ry', 'rz'])
            gates.append((d, g, rng.normal(1.0, 0.5), q))

    circ = qtn.Circuit(L)
    circ.apply_gates(gates)

    return circ


def qft_circ(n, swaps=True, **circuit_opts):

    circ = qtn.Circuit(n, **circuit_opts)

    for i in range(n):
        circ.h(i)
        for j, m in zip(range(i + 1, n), itertools.count(2)):
            circ.cu1(2 * math.pi / 2**m, j, i)

    if swaps:
        for i in range(n // 2):
            circ.swap(i, n - i - 1)

    return circ


def swappy_circ(n, depth):
    circ = qtn.Circuit(n)

    for d in range(depth):
        pairs = np.random.permutation(np.arange(n))

        for i in range(n // 2):
            qi = pairs[2 * i]
            qj = pairs[2 * i + 1]

            gate = np.random.choice(['FSIM', 'SWAP'])
            if gate == 'FSIM':
                params = np.random.randn(2)
            elif gate == 'FSIMG':
                params = np.random.randn(5)
            else:
                params = ()

            circ.apply_gate(gate, *params, qi, qj)

    return circ


class TestCircuit:

    def test_prepare_GHZ(self):
        qc = qtn.Circuit(3)
        gates = [
            ('H', 0),
            ('H', 1),
            ('CNOT', 1, 2),
            ('CNOT', 0, 2),
            ('H', 0),
            ('H', 1),
            ('H', 2),
        ]
        qc.apply_gates(gates)
        assert qu.expec(qc.psi.to_dense(), qu.ghz_state(3)) == pytest.approx(1)
        counts = qc.simulate_counts(1024)
        assert len(counts) == 2
        assert '000' in counts
        assert '111' in counts
        assert counts['000'] + counts['111'] == 1024

    def test_from_qasm(self):
        G = rand_reg_graph(reg=3, n=18, seed=42)
        qasm = graph_to_qasm(G)
        qc = qtn.Circuit.from_qasm(qasm)
        assert (qc.psi.H & qc.psi) ^ all == pytest.approx(1.0)

    def test_from_qasm_mps_swapsplit(self):
        G = rand_reg_graph(reg=3, n=18, seed=42)
        qasm = graph_to_qasm(G)
        qc = qtn.CircuitMPS.from_qasm(qasm)
        assert len(qc.psi.tensors) == 18
        assert (qc.psi.H & qc.psi) ^ all == pytest.approx(1.0)

    @pytest.mark.parametrize(
        'Circ', [qtn.Circuit, qtn.CircuitMPS, qtn.CircuitDense]
    )
    def test_all_gate_methods(self, Circ):
        import random

        g_nq_np = [
            # single qubit
            ('x', 1, 0),
            ('y', 1, 0),
            ('z', 1, 0),
            ('s', 1, 0),
            ('t', 1, 0),
            ('h', 1, 0),
            ('iden', 1, 0),
            ('x_1_2', 1, 0),
            ('y_1_2', 1, 0),
            ('z_1_2', 1, 0),
            ('w_1_2', 1, 0),
            ('hz_1_2', 1, 0),
            # single qubit parametrizable
            ('rx', 1, 1),
            ('ry', 1, 1),
            ('rz', 1, 1),
            ('u3', 1, 3),
            ('u2', 1, 2),
            ('u1', 1, 1),
            # two qubit
            ('cx', 2, 0),
            ('cy', 2, 0),
            ('cz', 2, 0),
            ('cnot', 2, 0),
            ('swap', 2, 0),
            ('iswap', 2, 0),
            # two qubit parametrizable
            ('cu3', 2, 3),
            ('cu2', 2, 2),
            ('cu1', 2, 1),
            ('fsim', 2, 2),
            ('fsimg', 2, 5),
            ('rzz', 2, 1),
            ('su4', 2, 15),
        ]
        random.shuffle(g_nq_np)

        psi0 = qtn.MPS_rand_state(2, 2)
        circ = Circ(2, psi0, tags='PSI0')

        for g, n_q, n_p in g_nq_np:
            args = [
                *np.random.uniform(0, 2 * np.pi, size=n_p),
                *np.random.choice([0, 1], replace=False, size=n_q)
            ]
            getattr(circ, g)(*args)

        assert circ.psi.H @ circ.psi == pytest.approx(1.0)
        assert abs((circ.psi.H & psi0) ^ all) < 0.99999999

    def test_su4(self):
        psi0 = qtn.MPS_rand_state(2, 2)
        circ_a = qtn.Circuit(psi0=psi0)
        params = qu.randn(15)

        circ_a.su4(*params, 0, 1)
        psi_a = circ_a.to_dense()

        circ_b = qtn.Circuit(psi0=psi0)
        (theta1, phi1, lamda1,
         theta2, phi2, lamda2,
         theta3, phi3, lamda3,
         theta4, phi4, lamda4,
         t1, t2, t3,) = params
        circ_b.u3(theta1, phi1, lamda1, 0)
        circ_b.u3(theta2, phi2, lamda2, 1)
        circ_b.cnot(1, 0)
        circ_b.rz(t1, 0)
        circ_b.ry(t2, 1)
        circ_b.cnot(0, 1)
        circ_b.ry(t3, 1)
        circ_b.cnot(1, 0)
        circ_b.u3(theta3, phi3, lamda3, 0)
        circ_b.u3(theta4, phi4, lamda4, 1)
        psi_b = circ_b.to_dense()

        assert qu.fidelity(psi_a, psi_b) == pytest.approx(1.0)

    def test_auto_split_gate(self):

        n = 3
        ops = [
            ('u3', 1., 2., 3., 0),
            ('u3', 2., 3., 1., 1),
            ('u3', 3., 1., 2., 2),
            ('cz', 0, 1),
            ('iswap', 1, 2),
            ('cx', 2, 0),
            ('iswap', 2, 1),
            ('h', 0),
            ('h', 1),
            ('h', 2),
        ]
        cnorm = qtn.Circuit(n, gate_opts=dict(contract='split-gate'))
        cnorm.apply_gates(ops)
        assert cnorm.psi.max_bond() == 4

        cswap = qtn.Circuit(n, gate_opts=dict(contract='swap-split-gate'))
        cswap.apply_gates(ops)
        assert cswap.psi.max_bond() == 4

        cauto = qtn.Circuit(n, gate_opts=dict(contract='auto-split-gate'))
        cauto.apply_gates(ops)
        assert cauto.psi.max_bond() == 2

        assert qu.fidelity(cnorm.psi.to_dense(),
                           cswap.psi.to_dense()) == pytest.approx(1.0)
        assert qu.fidelity(cswap.psi.to_dense(),
                           cauto.psi.to_dense()) == pytest.approx(1.0)

    @pytest.mark.parametrize("gate2", ['cx', 'iswap'])
    def test_circuit_simplify_tensor_network(self, gate2):
        import random
        import itertools

        depth = n = 8

        circ = qtn.Circuit(n)

        def random_single_qubit_layer():
            return [
                (random.choice(['X_1_2', 'Y_1_2', 'W_1_2']), i)
                for i in range(n)
            ]

        def even_two_qubit_layer():
            return [
                (gate2, i, i + 1)
                for i in range(0, n, 2)
            ]

        def odd_two_qubit_layer():
            return [
                (gate2, i, i + 1)
                for i in range(1, n - 1, 2)
            ]

        layering = itertools.cycle([
            random_single_qubit_layer,
            even_two_qubit_layer,
            random_single_qubit_layer,
            odd_two_qubit_layer,
        ])

        for i, layer_fn in zip(range(depth), layering):
            for g in layer_fn():
                circ.apply_gate(*g, gate_round=i)

        psif = qtn.MPS_computational_state('0' * n).squeeze_()
        tn = circ.psi & psif

        c = tn.contract(all)
        cw = tn.contraction_width()

        tn_s = tn.full_simplify()
        assert tn_s.num_tensors < tn.num_tensors
        assert tn_s.num_indices < tn.num_indices
        # need to specify output inds since we now have hyper edges
        c_s = tn_s.contract(all, output_inds=[])
        assert c_s == pytest.approx(c)
        cw_s = tn_s.contraction_width(output_inds=[])
        assert cw_s <= cw

    def test_amplitude(self):
        L = 5
        circ = random_a2a_circ(L, 3)
        psi = circ.to_dense()

        for i in range(2**L):
            b = f"{i:0>{L}b}"
            c = circ.amplitude(b)
            assert c == pytest.approx(psi[i, 0])

    def test_partial_trace(self):
        L = 5
        circ = random_a2a_circ(L, 3)
        psi = circ.to_dense()
        for i in range(L - 1):
            keep = (i, i + 1)
            assert_allclose(qu.partial_trace(psi, [2] * 5, keep=keep),
                            circ.partial_trace(keep),
                            atol=1e-12)

    @pytest.mark.parametrize("group_size", (1, 2, 6))
    def test_sample(self, group_size):
        import collections
        from scipy.stats import power_divergence

        C = 2**10
        L = 5
        circ = random_a2a_circ(L, 3)

        psi = circ.to_dense()
        p_exp = abs(psi.reshape(-1))**2
        f_exp = p_exp * C

        counts = collections.Counter(circ.sample(C, group_size=group_size))
        f_obs = np.zeros(2**L)
        for b, c in counts.items():
            f_obs[int(b, 2)] = c

        assert power_divergence(f_obs, f_exp)[0] < 100

    def test_sample_chaotic(self):
        import collections
        from scipy.stats import power_divergence

        C = 2**12
        L = 5
        reps = 3
        depth = 2
        goodnesses = [0] * 5

        for _ in range(reps):
            circ = random_a2a_circ(L, depth)

            psi = circ.to_dense()
            p_exp = abs(psi.reshape(-1))**2
            f_exp = p_exp * C

            for num_marginal in [3, 4, 5]:
                counts = collections.Counter(
                    circ.sample_chaotic(C, num_marginal, seed=666)
                )
                f_obs = np.zeros(2**L)
                for b, c in counts.items():
                    f_obs[int(b, 2)] = c

                goodness = power_divergence(f_obs, f_exp)[0]
                goodnesses[num_marginal - 1] += goodness

        # assert average sampling goodness gets better with larger marginal
        assert sum(goodnesses[i] < goodnesses[i - 1] for i in range(1, L)) == 2

    def test_local_expectation(self):
        import random
        L = 5
        depth = 3
        circ = random_a2a_circ(L, depth)
        psi = circ.to_dense()
        for _ in range(10):
            G = qu.rand_matrix(4)
            i = random.randint(0, L - 2)
            where = (i, i + 1)
            x1 = qu.expec(qu.ikron(G, [2] * L, where), psi)
            x2 = circ.local_expectation(G, where)
            assert x1 == pytest.approx(x2)

    def test_local_expectation_multigate(self):
        circ = qtn.Circuit(2)
        circ.h(0)
        circ.cnot(0, 1)
        circ.y(1)
        Gs = [qu.kronpow(qu.pauli(s), 2) for s in 'xyz']
        exps = circ.local_expectation(Gs, [0, 1])
        assert exps[0] == pytest.approx(-1)
        assert exps[1] == pytest.approx(-1)
        assert exps[2] == pytest.approx(-1)

    def test_uni_to_dense(self):
        import cmath
        circ = qft_circ(3)
        U = circ.uni.to_dense()
        w = cmath.exp(2j * math.pi / 2**3)
        ex = 2**(-3 / 2) * np.array(
            [[w**0, w**0, w**0, w**0, w**0, w**0, w**0, w**0],
             [w**0, w**1, w**2, w**3, w**4, w**5, w**6, w**7],
             [w**0, w**2, w**4, w**6, w**0, w**2, w**4, w**6],
             [w**0, w**3, w**6, w**1, w**4, w**7, w**2, w**5],
             [w**0, w**4, w**0, w**4, w**0, w**4, w**0, w**4],
             [w**0, w**5, w**2, w**7, w**4, w**1, w**6, w**3],
             [w**0, w**6, w**4, w**2, w**0, w**6, w**4, w**2],
             [w**0, w**7, w**6, w**5, w**4, w**3, w**2, w**1]])
        assert_allclose(U, ex)

    def test_swap_lighcones(self):
        circ = qtn.Circuit(3)
        circ.x(0)  # 0
        circ.x(1)  # 1
        circ.x(2)  # 2
        circ.swap(0, 1)  # 3
        circ.cx(1, 2)  # 4
        circ.cx(0, 1)  # 5
        assert circ.get_reverse_lightcone_tags((2,)) == (
            'PSI0', 'GATE_0', 'GATE_2', 'GATE_4'
        )

    def test_swappy_local_expecs(self):
        circ = swappy_circ(4, 4)
        Gs = [qu.rand_matrix(4) for _ in range(3)]
        pairs = [(0, 1), (1, 2), (2, 3)]

        psi = circ.to_dense()
        dims = [2] * 4

        exs = [qu.expec(qu.ikron(G, dims, pair), psi)
               for G, pair in zip(Gs, pairs)]
        aps = [circ.local_expectation(G, pair) for G, pair in zip(Gs, pairs)]

        assert_allclose(exs, aps)

    @pytest.mark.parametrize(
        "name, densefn, nparam, nqubit",
        [
            ('rx', qu.Rx, 1, 1),
            ('ry', qu.Ry, 1, 1),
            ('rz', qu.Rz, 1, 1),
            ('u3', qu.U_gate, 3, 1),
            ('fsim', qu.fsim, 2, 2),
            ('fsimg', qu.fsimg, 5, 2),
        ]
    )
    def test_parametrized_gates_rx(self, name, densefn, nparam, nqubit):
        k0 = qu.rand_ket(2**nqubit)
        params = qu.randn(nparam)
        kf = densefn(*params) @ k0
        k0mps = qtn.MatrixProductState.from_dense(k0, [2] * nqubit)
        circ = qtn.Circuit(psi0=k0mps, gate_opts={'contract': False})
        getattr(circ, name)(*params, *range(nqubit), parametrize=True)
        tn = circ.psi
        assert isinstance(tn['GATE_0'], qtn.PTensor)
        assert_allclose(circ.to_dense(), kf)

    def test_apply_raw_gate(self):
        k0 = qu.rand_ket(4)
        psi0 = qtn.MatrixProductState.from_dense(k0, [2] * 2)
        circ = qtn.Circuit(psi0=psi0)
        U = qu.rand_uni(4)
        circ.apply_gate_raw(U, [0, 1], tags='UCUSTOM')
        assert len(circ.gates) == 1
        assert 'UCUSTOM' in circ.psi.tags
        assert qu.fidelity(circ.to_dense(), U @ k0) == pytest.approx(1)


class TestCircuitGen:

    @pytest.mark.parametrize(
        "ansatz,cyclic", [
            ('zigzag', False),
            ('brickwork', False),
            ('brickwork', True),
            ('rand', False),
            ('rand', True),
        ])
    @pytest.mark.parametrize('n', [4, 5])
    def test_1D_ansatzes(self, ansatz, cyclic, n):
        depth = 3
        num_pairs = n if cyclic else n - 1

        fn = {
            'zigzag': qtn.circ_ansatz_1D_zigzag,
            'brickwork': qtn.circ_ansatz_1D_brickwork,
            'rand': qtn.circ_ansatz_1D_rand,
        }[ansatz]

        opts = dict(
            n=n,
            depth=3,
            gate_opts=dict(contract=False),
        )
        if cyclic:
            opts['cyclic'] = True
        if ansatz == 'rand':
            opts['seed'] = 42

        circ = fn(**opts)
        tn = circ.uni

        # total number of entangling gates
        assert len(tn['CZ']) == num_pairs * depth

        # number of entangling gates per pair
        for i in range(num_pairs):
            assert len(tn['CZ', f'I{i}', f'I{(i + 1) % n}']) == depth

        assert all(isinstance(t, qtn.PTensor) for t in tn['U3'])

    def test_qaoa(self):
        G = rand_reg_graph(3, 10, seed=666)
        terms = {(i, j): 1. for i, j in G.edges}
        ZZ = qu.pauli('Z') & qu.pauli('Z')

        gammas = [1.2]
        betas = [-0.4]

        circ1 = qtn.circ_qaoa(terms, 1, gammas, betas)

        energy1 = sum(
            circ1.local_expectation(ZZ, edge)
            for edge in terms
        )
        assert energy1 < -4

        gammas = [0.4]
        betas = [0.3]

        circ2 = qtn.circ_qaoa(terms, 1, gammas, betas)

        energy2 = sum(
            circ2.local_expectation(ZZ, edge)
            for edge in terms
        )
        assert energy2 > 4



"""Tools for defining and constructing sparse operators with:

    * arbitrary geometries,
    * numba acceleration,
    * support for reduced bases,
    * efficient parallelization,

and optionally producing:

    * sparse matrix form
    * matrix product operators,
    * dict of local gates form
    * VMC 'coupled configs' form

Currently only supports composing operators which are sums of products of
diagonal or anti-diagonal real dimension 2 operators.

TODO:

    - [ ] automatic MPO generator
    - [ ] product of operators generator (e.g. for PEPS DMRG)
    - [ ] complex and single precision support (lower priority)
    - [ ] support for non-diagonal and dimension > 2 operators (lower priority)

DONE:

    - [x] use compact bitbasis
    - [x] design interface for HilbertSpace / OperatorBuilder interaction
    - [x] automatic symbolic jordan wigner transformation
    - [x] numba accelerated coupled config
    - [x] general definition and automatic 'symbolic' jordan wigner
    - [x] multithreaded sparse matrix construction
    - [x] LocalHam generator (e.g. for simple update, normal PEPS algs)

"""

import functools
import collections

import numpy as np
from numba import njit


@njit
def get_local_size(n, rank, world_size):
    """Given global size n, and a rank in [0, world_size), return the size of
    the portion assigned to this rank.
    """
    cutoff_rank = n % world_size
    return n // world_size + int(rank < cutoff_rank)


@njit
def get_local_range(n, rank, world_size):
    """Given global size n, and a rank in [0, world_size), return the range of
    indices assigned to this rank.
    """
    ri = 0
    for rank_below in range(rank):
        ri += get_local_size(n, rank_below, world_size)
    rf = ri + get_local_size(n, rank, world_size)
    return ri, rf


def parse_edges_to_unique(edges):
    """Given a list of edges, return a sorted list of unique sites and edges.

    Parameters
    ----------
    edges : sequence of (hashable, hashable)
        The edges to parse.

    Returns
    -------
    sites : list of hashable
        The unique sites in the edges, sorted.
    edges : list of (hashable, hashable)
        The unique edges, sorted.
    """
    sites = set()
    uniq_edges = set()
    for i, j in edges:
        if j < i:
            i, j = j, i
        sites.add(i)
        sites.add(j)
        uniq_edges.add((i, j))
    return sorted(sites), sorted(uniq_edges)


class HilbertSpace:
    """Take a set of 'sites' (any sequence of sortable, hashable objects), and
    map this into a 'register' or linearly indexed range, optionally using a
    particular ordering. One can then calculate the size of the Hilbert space,
    including number conserving subspaces, and get a compact 'bitbasis' with
    which to construct sparse operators.

    Parameters
    ----------
    sites : int or sequence of hashable objects
        The sites to map into a linear register. If an integer, simply use
        ``range(sites)``.
    order : callable or sequence of hashable objects, optional
        If provided, use this to order the sites. If a callable, it should be a
        sorting key. If a sequence, it should be a permutation of the sites,
        and ``key=order.index`` will be used.
    """

    def __init__(self, sites, order=None):

        if isinstance(sites, int):
            sites = range(sites)

        if (order is not None) and (not callable(order)):
            order = order.index
        self._order = order
        self._sites = tuple(sorted(sites, key=self._order))
        self._mapping_inv = dict(enumerate(self._sites))
        self._mapping = {s: i for i, s in self._mapping_inv.items()}

    @classmethod
    def from_edges(cls, edges, order=None):
        """Construct a HilbertSpace from a set of edges, which are pairs of
        sites.
        """
        sites, _ = parse_edges_to_unique(edges)
        return cls(sites, order=order)

    def site_to_reg(self, site):
        """Convert a site to a linear register index."""
        return self._mapping[site]

    def reg_to_site(self, reg):
        """Convert a linear register index back to a site."""
        return self._mapping_inv[reg]

    def has_site(self, site):
        """Check if this HilbertSpace contains a given site."""
        return site in self._mapping

    def config_to_bit(self, config):
        """Encode a 'configuration' as a bit.

        Parameters
        ----------
        config : dict[hashable, int]
            A dictionary mapping sites to their occupation number / spin.

        Returns
        -------
        bit : int
            The bit corresponding to this configuration.
        """
        bit = 0
        for site, val in config.items():
            if val:
                bit |= 1 << self.site_to_reg(site)
        return bit

    def config_to_flatconfig(self, config):
        """Turn a configuration into a flat configuration, assuming the order
        given by this ``HilbertSpace``.

        Parameters
        ----------
        config : dict[hashable, int]
            A dictionary mapping sites to their occupation number / spin.

        Returns
        -------
        flatconfig : ndarray[uint8]
            A flat configuration, with the occupation number of each site in
            the order given by this ``HilbertSpace``.
        """
        flatconfig = np.empty(self.nsites, dtype=np.uint8)
        for site, val in config.items():
            flatconfig[self.site_to_reg(site)] = val
        return flatconfig

    def bit_to_config(self, bit):
        """Decode a bit to a configuration.

        Parameters
        ----------
        bit : int
            The bit to decode.

        Returns
        -------
        config : dict[hashable, int]
            A dictionary mapping sites to their occupation number / spin.
        """
        config = {}
        for site, reg in self._mapping.items():
            config[site] = (bit >> reg) & 1
        return config

    def flatconfig_to_config(self, flatconfig):
        """Turn a flat configuration into a configuration, assuming the order
        given by this ``HilbertSpace``.

        Parameters
        ----------
        flatconfig : ndarray[uint8]
            A flat configuration, with the occupation number of each site in
            the order given by this ``HilbertSpace``.

        Returns
        -------
        config : dict[hashable, int]
            A dictionary mapping sites to their occupation number / spin.
        """
        config = {}
        for i in range(self.nsites):
            config[self.reg_to_site(i)] = flatconfig[i]
        return config

    @property
    def sites(self):
        """The ordered tuple of all sites in the Hilbert space."""
        return self._sites

    @property
    def nsites(self):
        """The total number of sites in the Hilbert space."""
        return len(self._sites)

    def get_size(self, *k):
        """Compute the size of this Hilbert space, optionally taking into
        account number / z-spin conservation.

        Parameters
        ----------
        k : int or tuple of (int, int)
            If provided, compute the size of number conserving subspace(s)::

                - If a single int, compute the size of the subspace with
                  ``k`` particles / up states: ``comb(nsites, k)``.
                - If a tuple of (int, int), compute the size of the subspace
                  of the product of spaces where each pair (n, k) corresponds
                  to n sites with k particles / up states. The sum of every n
                  should equal ``nsites``.
        """
        if not k:
            return 2**self.nsites

        if (len(k) == 1) and isinstance(k[0], int):
            # single interger - take as k
            return comb(self.nsites, k[0])

        size = 1
        ncheck = 0
        for n, k in k:
            size *= comb(n, k)
            ncheck += n

        if ncheck != self.nsites:
            raise ValueError("`k` must sum to the number of sites")

        return size

    def rand_config(self, k=None):
        """Get a random configuration, optionally requiring it has ``k`` bits
        set.
        """
        if k is None:
            return {
                site: np.random.randint(2)
                for site in self.sites
            }
        r = np.random.randint(self.get_size(k))
        b = rank_to_bit(r, self.nsites, k)
        return self.bit_to_config(b)

    def get_bitbasis(self, *k, dtype=np.int64):
        """Get a basis for the Hilbert space, in terms of an integer bitarray,
        optionally taking into account number / z-spin conservation.

        Parameters
        ----------
        k : int or tuple of (int, int)
            If provided, get the basis for a number conserving subspace(s)::

                - If a single int, compute the size of the subspace with
                  ``k`` particles / up states: ``comb(nsites, k)``.
                - If a tuple of (int, int), compute the size of the subspace
                  of the product of spaces where each pair (n, k) corresponds
                  to n sites with k particles / up states. The sum of every n
                  should equal ``nsites``.

        dtype : np.dtype, optional
            The dtype of the bitarray, should be an integer type with at least
            ``nsites`` bits.

        Returns
        -------
        bits : numpy.ndarray
            The basis, each integer element being a binary representation of a
            configuration.
        """
        if not k:
            return np.arange(1 << self.nsites, dtype=dtype)

        if (len(k) == 1) and isinstance(k[0], int):
            k = ((self.nsites, k[0]),)

        return get_number_bitbasis(*k, dtype=dtype)

    def __repr__(self):
        return (
            f"HilbertSpace(nsites={self.nsites}, total_size={self.get_size()})"
        )


_OPMAP = {
    # pauli matrices
    "x": {0: (1, 1.0), 1: (0, 1.0)},
    "y": {0: (1, 1.0j), 1: (0, -1.0j)},
    "z": {0: (0, 1.0), 1: (1, -1.0)},
    # spin 1/2 matrices (scaled paulis)
    "sx": {0: (1, 0.5), 1: (0, 0.5)},
    "sy": {0: (1, 0.5j), 1: (0, -0.5j)},
    "sz": {0: (0, 0.5), 1: (1, -0.5)},
    # creation / annihilation operators
    "+": {0: (1, 1.0)},
    "-": {1: (0, 1.0)},
    # number and symmetric number operator
    "n": {1: (1, 1.0)},
    "sn": {0: (0, -0.5), 1: (1, 0.5)},
}


@functools.lru_cache(maxsize=None)
def get_mat(op, dtype=None):
    if dtype is None:
        if op in {'y', 'sy'}:
            dtype = np.complex128
        else:
            dtype = np.float64
    a = np.zeros((2, 2), dtype=dtype)
    for j, (i, xij) in _OPMAP[op].items():
        a[i, j] = xij
    # make immutable since caching
    a.flags.writeable = False
    return a


class SparseOperatorBuilder:
    """Object for building operators with sparse structure. Specifically,
    a sum of terms, where each term is a product of operators, where each of
    these local operators acts on a single site and has at most one entry per
    row.

    Parameters
    ----------
    terms : sequence, optional
        The terms to initialize the builder with. ``add_term`` is simply called
        on each of these.
    hilbert_space : HilbertSpace
        The Hilbert space to build the operator in. If this is not supplied
        then a minimal Hilbert space will be constructed from the sites used.
    """

    def __init__(self, terms=(), hilbert_space=None):
        self._terms = []
        self._sites_used = set()
        self._hilbert_space = hilbert_space
        self._coupling_map = None
        for coeff, ops, *sites in terms:
            self.add_term(coeff, ops, *sites)

    @property
    def sites_used(self):
        """A tuple of the sorted coordinates/sites seen so far."""
        return tuple(sorted(self._sites_used))

    @property
    def nsites(self):
        """The total number of coordinates/sites seen so far."""
        return self.hilbert_space.nsites

    @property
    def terms(self):
        """A tuple of the terms seen so far."""
        return tuple(self._terms)

    @property
    def nterms(self):
        """The total number of terms seen so far."""
        return len(self._terms)

    @property
    def locality(self):
        """The locality of the operator, the maximum support of any term."""
        return max(len(sites) for _, _, *sites in self.terms)

    @property
    def hilbert_space(self):
        """The Hilbert space of the operator. Created from the sites seen
        so far if not supplied at construction.
        """
        if self._hilbert_space is None:
            self._hilbert_space = HilbertSpace(self.sites_used)
        return self._hilbert_space

    @property
    def coupling_map(self):
        if self._coupling_map is None:
            self._coupling_map = build_coupling_numba(
                self.terms, self.hilbert_space.site_to_reg
            )
        return self._coupling_map

    def site_to_reg(self, site):
        """Get the register / linear index of coordinate ``site``."""
        return self.hilbert_space.site_to_reg(site)

    def reg_to_site(self, reg):
        """Get the site of register / linear index ``reg``."""
        return self.hilbert_space.reg_to_site(reg)

    def add_term(self, coeff, ops, *sites):
        """Add a term to the operator.

        Parameters
        ----------
        coeff : float
            The overall coefficient of the term.
        ops : sequence of str
            The operators of the term, that together form a product. Each
            ``op`` should be one of:

                - ``'x'``, ``'y'``, ``'z'``: Pauli matrices
                - ``'sx'``, ``'sy'``, ``'sz'``: spin operators (i.e. scaled
                  Pauli matrices)
                - ``'+'``, ``'-'``: creation/annihilation operators
                - ``'n'``, ``'sn'``: number and symmetric number (n - 1/2)
                   operators

        sites : sequence of hashable
            The coordinates of the operators that ``ops`` acts on.
        """
        self._terms.append((coeff, tuple(ops), *sites))
        for site in sites:
            if (
                (self._hilbert_space is not None) and not
                self._hilbert_space.has_site(site)
            ):
                raise ValueError(f"Site {site} not in the Hilbert space.")
            self._sites_used.add(site)

    def __iadd__(self, term):
        self.add_term(*term)
        return self

    def jordan_wigner_transform(self):
        """Transform the terms in this operator by pre-prending pauli Z
        strings to all creation and annihilation operators, and then
        simplifying the resulting terms.
        """
        new_terms = []

        for coeff, old_ops, *sites in self._terms:
            if {"+", "-"}.intersection(old_ops):

                # explicitly collect and simplify z strings:
                jw_strings = collections.defaultdict(str)
                for op, site in zip(old_ops, sites):
                    # get the linear index of the site
                    reg = self.site_to_reg(site)
                    if op in {"+", "-"}:
                        # apply z everywhere below site
                        for i in range(reg):
                            jw_strings[i] += "z"
                    # terminate with the actual operator
                    jw_strings[reg] += op

                # now simplify the z strings
                new_ops, new_sites = [], []
                for reg, jw_strings in jw_strings.items():
                    jw_simplified = (
                        jw_strings.replace("zz", "")
                        .replace("+z", "+")
                        .replace("z-", "-")
                    )
                    if jw_simplified:
                        # keep only non identity operator terms
                        new_ops.append(jw_simplified)
                        new_sites.append(self.reg_to_site(reg))

                # replace term with the new one, strings inserted
                new_terms.append((coeff, tuple(new_ops), *new_sites))
            else:
                # doesn't involve any creation/annihilation operators
                new_terms.append((coeff, old_ops, *sites))

        self._terms = new_terms

    def build_coo_data(self, *k, parallel=False):
        """Build the raw data for a sparse matrix in COO format. Optionally
        in a reduced k basis and in parallel.

        Parameters
        ----------
        k : int or tuple of (int, int)
            If provided, get the basis for a number conserving subspace(s)::

                - If a single int, compute the size of the subspace with
                  ``k`` particles / up states: ``comb(nsites, k)``.
                - If a tuple of (int, int), compute the size of the subspace
                  of the product of spaces where each pair (n, k) corresponds
                  to n sites with k particles / up states. The sum of every n
                  should equal ``nsites``.

        parallel : bool, optional
            Whether to build the matrix in parallel (multi-threaded).

        Returns
        -------
        coo : array
            The data entries for the sparse matrix in COO format.
        cis : array
            The row indices for the sparse matrix in COO format.
        cjs : array
            The column indices for the sparse matrix in COO format.
        N : int
            The total number of basis states.
        """
        hs = self.hilbert_space
        bits = hs.get_bitbasis(*k)
        coupling_map = self.coupling_map
        coo, cis, cjs = build_coo_numba(bits, coupling_map, parallel=parallel)
        return coo, cis, cjs, bits.size

    def build_sparse_matrix(self, *k, stype="csr", parallel=False):
        """Build a sparse matrix in the given format. Optionally in a reduced
        k basis and in parallel.

        Parameters
        ----------
        k : int or tuple of (int, int)
            If provided, get the basis for a number conserving subspace(s)::

                - If a single int, compute the size of the subspace with
                  ``k`` particles / up states: ``comb(nsites, k)``.
                - If a tuple of (int, int), compute the size of the subspace
                  of the product of spaces where each pair (n, k) corresponds
                  to n sites with k particles / up states. The sum of every n
                  should equal ``nsites``.

        parallel : bool, optional
            Whether to build the matrix in parallel (multi-threaded).

        Returns
        -------
        scipy.sparse matrix
        """
        import scipy.sparse as sp
        coo, cis, cjs, N = self.build_coo_data(*k, parallel=parallel)
        A = sp.coo_matrix((coo, (cis, cjs)), shape=(N, N))
        if stype != "coo":
            A = A.asformat(stype)
        return A

    def build_dense(self):
        """Get the dense (`numpy.ndarray`) matrix representation of this
        operator.
        """
        A = self.build_sparse_matrix(stype="coo")
        return A.A

    def build_local_terms(self):
        """Get a dictionary of local terms, where each key is a sorted tuple
        of sites, and each value is the local matrix representation of the
        operator on those sites. For use with e.g. tensor network algorithms.

        Note terms acting on the same sites are summed together and the size of
        each local matrix is exponential in the locality of that term.

        Returns
        -------
        Hk : dict[tuple[hashable], numpy.ndarray]
            The local terms.
        """
        Hk = {}
        for coeff, ops, *sites in self.terms:
            # sort ops and sites according to sites
            sites, ops = zip(*sorted(zip(sites, ops)))
            mats = tuple(get_mat(op) for op in ops)
            hk = coeff * functools.reduce(np.kron, mats)
            if sites not in Hk:
                Hk[sites] = hk.copy()
            else:
                Hk[sites] += hk
        return Hk

    def config_coupling(self, config):
        """Get a list of other configurations coupled to ``config`` by this
        operator, and the corresponding coupling coefficients. This is for
        use with VMC for example.

        Parameters
        ----------
        config : dict[site, int]
            The configuration to get the coupling for.

        Returns
        -------
        coupled_configs : list[dict[site, int]]
            Each distinct configuration coupled to ``config``.
        coeffs: list[float]
            The corresponding coupling coefficients.
        """
        bit_to_config = self.hilbert_space.bit_to_config
        config_to_bit = self.hilbert_space.config_to_bit
        b = config_to_bit(config)
        bjs, coeffs = coupled_bits_numba(b, self.coupling_map)
        return [bit_to_config(bj) for bj in bjs], coeffs

    def show(self):
        """Print an ascii representation of the terms in this operator.
        """
        print(self)
        for t, (coeff, ops, *sites) in enumerate(self.terms):
            s = ['. '] * self.nsites
            for op, site in zip(ops, sites):
                s[self.site_to_reg(site)] = f"{op:<2}"
            print("".join(s), f"{coeff:+}")

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(nsites={self.nsites}, "
            f"nterms={self.nterms}, "
            f"locality={self.locality})"
            ")"
        )


@njit
def get_nth_bit(val, n):
    """Get the nth bit of val.

    Examples
    --------

        >>> get_nth_bit(0b101, 1)
        0
    """
    return (val >> n) & 1


@njit
def flip_nth_bit(val, n):
    """Flip the nth bit of val.

    Examples
    --------

        >>> bin(flip_nth_bit(0b101, 1))
        0b111
    """
    return val ^ (1 << n)


@njit
def comb(n, k):
    """Compute the binomial coefficient n choose k."""
    r = 1
    for dk, dn in zip(range(1, k + 1), range(n, n - k, -1)):
        r *= dn
        r //= dk
    return r


@njit
def get_all_equal_weight_bits(n, k, dtype=np.int64):
    """Get an array of all 'bits' (integers), with n bits, and k of them set."""
    if k == 0:
        return np.array([0], dtype=dtype)

    m = comb(n, k)
    b = np.empty(m, dtype=dtype)
    i = 0
    val = (1 << k) - 1
    while val < (1 << (n)):
        b[i] = val
        c = val & -val
        r = val + c
        val = (((r ^ val) >> 2) // c) | r
        i += 1
    return b


@njit
def bit_to_rank(b, n, k):
    """Given a bitstring b, return the rank of the bitstring in the
    basis of all bitstrings of length n with k bits set. Adapted from
    https://dlbeer.co.nz/articles/kwbs.html.
    """
    c = comb(n, k)
    r = 0
    while n:
        c0 = c * (n - k) // n
        if (b >> n - 1) & 1:
            r += c0
            k -= 1
            c -= c0
        else:
            c = c0
        n -= 1
    return r


@njit
def rank_to_bit(r, n, k):
    """Given a rank r, return the bitstring of length n with k bits set
    that has rank r in the basis of all bitstrings of length n with k
    bits set. Adapted from https://dlbeer.co.nz/articles/kwbs.html.
    """
    b = 0
    c = comb(n, k)
    while n:
        c0 = c * (n - k) // n
        if r >= c0:
            b |= 1 << (n - 1)
            r -= c0
            k -= 1
            c -= c0
        else:
            c = c0
        n -= 1
    return b


@njit
def _recursively_fill_flatconfigs(flatconfigs, n, k, c, r):
    c0 = c * (n - k) // n
    # set the entries of the left binary subtree
    flatconfigs[r: r + c0, -n] = 0
    # set the entries of the right binary subtree
    flatconfigs[r + c0: r + c, -n] = 1
    if n > 1:
        # process each subtree recursively
        _recursively_fill_flatconfigs(flatconfigs, n - 1, k, c0, r)
        _recursively_fill_flatconfigs(flatconfigs, n - 1, k - 1, c - c0, r + c0)


@njit
def get_all_equal_weight_flatconfigs(n, k):
    """Get every flat configuration of length n with k bits set.
    """
    c = comb(n, k)
    flatconfigs = np.empty((c, n), dtype=np.uint8)
    _recursively_fill_flatconfigs(flatconfigs, n, k, c, 0)
    return flatconfigs


@njit
def flatconfig_to_bit(flatconfig):
    """Given a flat configuration, return the corresponding bitstring.
    """
    b = 0
    for i, x in enumerate(flatconfig):
        if x:
            b |= 1 << i
    return b

@njit
def flatconfig_to_rank(flatconfig, n, k):
    """Given a flat configuration ``flatconfig``, return the rank of the
    bitstring in the basis of all bitstrings of length ``n`` with ``k`` bits
    set. Adapted from https://dlbeer.co.nz/articles/kwbs.html.
    """
    c = comb(n, k)
    r = 0
    while n:
        c0 = c * (n - k) // n
        if flatconfig[-n]:
            r += c0
            k -= 1
            c -= c0
        else:
            c = c0
        n -= 1
    return r


@njit
def rank_to_flatconfig(r, n, k):
    """Given a rank ``r``, return the flat configuration of length ``n``
    with ``k`` bits set that has rank ``r`` in the basis of all bitstrings of
    length ``n`` with ``k`` bits set. Adapted from
    https://dlbeer.co.nz/articles/kwbs.html.
    """
    flatconfig = np.zeros(n, dtype=np.uint8)
    c = comb(n, k)
    while n:
        c0 = c * (n - k) // n
        if r >= c0:
            flatconfig[-n] = 1
            r -= c0
            k -= 1
            c -= c0
        else:
            c = c0
        n -= 1
    return flatconfig


@njit
def product_of_bits(b1, b2, n2, dtype=np.int64):
    """Get the outer product of two bit arrays.

    Parameters
    ----------
    b1, b2 : array
        The bit arrays to take the outer product of.
    n2 : int
        The number of bits in ``b2``
    """
    b = np.empty(len(b1) * len(b2), dtype=dtype)
    i = 0
    for x in b1:
        for y in b2:
            b[i] = (x << n2) | y
            i += 1
    return b


def get_number_bitbasis(*nk_pairs, dtype=np.int64):
    """Create a bit basis with number conservation.

    Parameters
    ----------
    nk_pairs : sequence of (int, int)
        Each element is a pair (n, k) where n is the number of bits, and k is
        the number of bits that are set. The basis will be the product of all
        supplied pairs.

    Returns
    -------
    basis : ndarray
        An array of integers, each representing a bit string. The size will be
        ``prod(comb(n, k) for n, k in nk_pairs)``.

    Examples
    --------

    A single number conserving basis:

        >>> for i, b in enumerate(get_number_bitbasis((4, 2))):
        >>>    print(f"{i}: {b:0>4b}")
        0: 0011
        1: 0101
        2: 0110
        3: 1001
        4: 1010
        5: 1100

    A product of two number conserving bases, e.g. n_up and n_down:

        >>> for b in get_number_bitbasis((3, 2), (3, 1)):
        >>>     print(f"{b:0>6b}")
        011001
        011010
        011100
        101001
        101010
        101100
        110001
        110010
        110100

    """
    configs = None
    for n, k in nk_pairs:
        next_configs = get_all_equal_weight_bits(n, k, dtype=dtype)
        if configs is None:
            configs = next_configs
        else:
            configs = product_of_bits(configs, next_configs, n, dtype=dtype)
    return configs


@njit
def build_bitmap(configs):
    """Build of map of bits to linear indices, suitable for use with numba."""
    return {b: i for i, b in enumerate(configs)}


def build_coupling_numba(terms, site_to_reg):
    """Create a sparse nested dictionary of how each term couples each
    local site configuration to which other local site configuration, and
    with what coefficient, suitable for use with numba.

    Parameters
    ----------
    terms : sequence of (coeff, ops, *sites)
        The terms of the operator.
    site_to_reg : callable
        A function that maps a site to a linear register index.

    Returns
    -------
    coupling_map : numba.typed.Dict
        A nested numba dictionary of the form
        ``{term: {reg: {bit_in: (bit_out, coeff), ...}, ...}, ...}``.
    """
    from numba.core import types
    from numba.typed import Dict

    ty_xj = types.Tuple((types.int64, types.float64))
    ty_xi = types.DictType(types.int64, ty_xj)
    ty_site = types.DictType(types.int64, ty_xi)
    coupling_map = Dict.empty(types.int64, ty_site)

    # for term t ...
    for t, (coeff, ops, *sites) in enumerate(terms):
        first = True
        # which couples sites with product of ops ...
        for op, site in zip(ops, sites):
            reg = site_to_reg(site)
            # -> bit `xi` at `reg` is coupled to `xj` with coeff `cij`
            #          : reg
            #     ...10010...    xi=0  ->
            #     ...10110...    xj=1  with coeff cij
            for xi, (xj, cij) in _OPMAP[op].items():
                if first:
                    # absorb overall coefficient into first coupling
                    cij = coeff * cij

                # populate just the term/reg/bit maps we need
                coupling_map.setdefault(
                    t, Dict.empty(types.int64, ty_xi)
                ).setdefault(reg, Dict.empty(types.int64, ty_xj))[xi] = (
                    xj,
                    cij,
                )
            first = False

    return coupling_map


def build_coupling(terms, site_to_reg):
    coupling_map = dict()
    # for term t ...
    for t, (coeff, ops, *sites) in enumerate(terms):
        first = True
        # which couples sites with product of ops ...
        for op, site in zip(ops, sites):
            reg = site_to_reg(site)
            # -> bit `xi` at `reg` is coupled to `xj` with coeff `cij`
            #          : reg
            #     ...10010...    xi=0  ->
            #     ...10110...    xj=1  with coeff cij
            for xi, (xj, cij) in _OPMAP[op].items():
                if first:
                    # absorb overall coefficient into first coupling
                    cij = coeff * cij

                # populate just the term/reg/bit maps we need
                coupling_map.setdefault(t, {}).setdefault(reg, {})[xi] = (xj, cij)
            first = False

    return coupling_map


@njit(nogil=True)
def coupled_flatconfigs_numba(flatconfig, coupling_map):
    """Get the coupled flat configurations for a given flat configuration
    and coupling map.

    Parameters
    ----------
    flatconfig : ndarray[uint8]
        The flat configuration to get the coupled configurations for.
    coupling_map : numba.typed.Dict
        A nested numba dictionary of the form
        ``{term: {reg: {bit_in: (bit_out, coeff), ...}, ...}, ...}``.

    Returns
    -------
    coupled_flatconfigs : ndarray[uint8]
        A list of coupled flat configurations, each with the corresponding
        coefficient.
    coeffs : ndarray[float64]
        The coefficients for each coupled flat configuration.
    """
    buf_ptr = 0
    bjs = np.empty((len(coupling_map), flatconfig.size), dtype=np.uint8)
    cijs = np.empty(len(coupling_map), dtype=np.float64)
    for coupling_t in coupling_map.values():
        cij = 1.0
        bj = flatconfig.copy()
        # bjs[buf_ptr, :] = flatconfig
        for reg, coupling_t_reg in coupling_t.items():
            xi = flatconfig[reg]
            if xi not in coupling_t_reg:
                # zero coupling - whole branch dead
                break
            # update coeff and config
            xj, cij = coupling_t_reg[xi]
            cij *= cij
            if xi != xj:
                bj[reg] = xj
        else:
            # no break - all terms survived
            bjs[buf_ptr, :] = bj
            cijs[buf_ptr] = cij
            buf_ptr += 1
    return bjs[:buf_ptr], cijs[:buf_ptr]


@njit(nogil=True)
def coupled_bits_numba(bi, coupling_map):
    buf_ptr = 0
    bjs = np.empty(len(coupling_map), dtype=np.int64)
    cijs = np.empty(len(coupling_map), dtype=np.float64)
    bitmap = {}

    for coupling_t in coupling_map.values():
        cij = 1.0
        bj = bi
        for reg, coupling_t_reg in coupling_t.items():
            xi = get_nth_bit(bi, reg)
            if xi not in coupling_t_reg:
                # zero coupling - whole branch dead
                break
            # update coeff and config
            xj, cij = coupling_t_reg[xi]
            cij *= cij
            if xi != xj:
                bj = flip_nth_bit(bj, reg)
        else:
            # no break - all terms survived
            try:
                # already seed this config - just add coeff
                loc = bitmap[bj]
                cijs[loc] += cij
            except:
                bjs[buf_ptr] = bj
                cijs[buf_ptr] = cij
                bitmap[bj] = buf_ptr
                buf_ptr += 1

    return bjs[:buf_ptr], cijs[:buf_ptr]


@njit(nogil=True)
def _build_coo_numba_core(bits, coupling_map, bitmap=None):
    # the bit map is needed if we only have a partial set of `bits`, which
    # might couple to other bits that are not in `bits` -> we need to look up
    # the linear register of the coupled bit
    if bitmap is None:
        bitmap = {bi: ci for ci, bi in enumerate(bits)}

    buf_size = len(bits)
    data = np.empty(buf_size, dtype=np.float64)
    cis = np.empty(buf_size, dtype=np.int64)
    cjs = np.empty(buf_size, dtype=np.int64)
    buf_ptr = 0

    for bi in bits:
        ci = bitmap[bi]
        for coupling_t in coupling_map.values():
            hij = 1.0
            bj = bi
            for reg, coupling_t_reg in coupling_t.items():
                xi = get_nth_bit(bi, reg)
                if xi not in coupling_t_reg:
                    # zero coupling - whole branch dead
                    break
                # update coeff and config
                xj, cij = coupling_t_reg[xi]
                hij *= cij
                if xi != xj:
                    bj = flip_nth_bit(bj, reg)
            else:
                # didn't break out of loop
                if buf_ptr >= buf_size:
                    # need to double our storage
                    data = np.concatenate((data, np.empty_like(data)))
                    cis = np.concatenate((cis, np.empty_like(cis)))
                    cjs = np.concatenate((cjs, np.empty_like(cjs)))
                    buf_size *= 2
                data[buf_ptr] = hij
                cis[buf_ptr] = ci
                cjs[buf_ptr] = bitmap[bj]
                buf_ptr += 1

    return data[:buf_ptr], cis[:buf_ptr], cjs[:buf_ptr]


def build_coo_numba(bits, coupling_map, parallel=False):
    """Build an operator in COO form, using the basis ``bits`` and the
     ``coupling_map``, optionally multhreaded.

    Parameters
    ----------
    bits : array
        An array of integers, each representing a bit string.
    coupling_map : Dict[int, Dict[int, Dict[int, Tuple[int, float]]]]
        A nested numba dictionary of couplings. The outermost key is the term
        index, the next key is the register, and the innermost key is the bit
        index. The value is a tuple of (coupled bit index, coupling
        coefficient).
    parallel : bool or int, optional
        Whether to parallelize the computation. If an integer is given, it
        specifies the number of threads to use.

    Returns
    -------
    data : array
        The non-zero elements of the operator.
    cis : array
        The row indices of the non-zero elements.
    cjs : array
        The column indices of the non-zero elements.
    """
    if not parallel:
        return _build_coo_numba_core(bits, coupling_map)

    from quimb import get_thread_pool

    if isinstance(parallel, int):
        n_thread_workers = parallel
    else:
        n_thread_workers = None

    pool = get_thread_pool(n_thread_workers)
    n_thread_workers = pool._max_workers

    # need a global mapping of bits to linear indices
    kws = dict(coupling_map=coupling_map, bitmap=build_bitmap(bits))

    # launch the threads! note we distribtue in cyclic fashion as the sparsity
    # can be concentrated in certain ranges and we want each thread to have
    # roughly the same amount of work to do
    fs = [
        pool.submit(_build_coo_numba_core, bits=bits[i::n_thread_workers], **kws)
        for i in range(n_thread_workers)
    ]

    # gather and concatenate the results (probably some memory overhead here)
    data = []
    cis = []
    cjs = []
    for f in fs:
        d, ci, cj = f.result()
        data.append(d)
        cis.append(ci)
        cjs.append(cj)

    data = np.concatenate(data)
    cis = np.concatenate(cis)
    cjs = np.concatenate(cjs)

    return data, cis, cjs


# -------------------------- specific hamiltonians -------------------------- #

def fermi_hubbard_from_edges(edges, t=1.0, U=1.0, mu=0.0):
    """
    """
    H = SparseOperatorBuilder()
    sites, edges = parse_edges_to_unique(edges)

    if t != 0.0:
        for cooa, coob in edges:
            # hopping
            for s in ['↑', '↓']:
                H += (-t, '+-', (s, *cooa), (s, *coob))
                H += (-t, '+-', (s, *coob), (s, *cooa))

    for coo in sites:
        # interaction
        if U != 0.0:
            H += (U, 'nn', ('↑', *coo), ('↓', *coo))

        # chemical potential
        if mu != 0.0:
            H += (mu, 'n', ('↑', *coo))
            H += (mu, 'n', ('↓', *coo))

    H.jordan_wigner_transform()
    return H


def fermi_hubbard_spinless_from_edges(edges, t=1.0, mu=0.0):
    H = SparseOperatorBuilder()
    sites, edges = parse_edges_to_unique(edges)

    for cooa, coob in edges:
        # hopping
        H += (-t, '+-', cooa, coob)
        H += (-t, '+-', coob, cooa)

    # chemical potential
    if mu != 0.0:
        for coo in sites:
            H += (mu, 'n', coo)

    H.jordan_wigner_transform()
    return H


def heisenberg_from_edges(edges, j=1.0, b=0.0, hilbert_space=None):
    """
    """
    try:
        jx, jy, jz = j
    except TypeError:
        jx, jy, jz = j, j, j

    try:
        bx, by, bz = b
    except TypeError:
        bx, by, bz = 0, 0, b

    H = SparseOperatorBuilder(hilbert_space=hilbert_space)
    sites, edges = parse_edges_to_unique(edges)

    for cooa, coob in edges:
        if (jx == jy):
            if (jx != 0.0):
                H += (jx / 2, '+-', cooa, coob)
                H += (jx / 2, '-+', cooa, coob)
        else:
            if jx != 0.0:
                H += (jx, ('sx', 'sx'), cooa, coob)
            if jy != 0.0:
                H += (jy, ('sy', 'sy'), cooa, coob)

        if jz != 0.0:
            H += (j, ('sz', 'sz'), cooa, coob)

    for site in sites:
        if bx != 0.0:
            H += (bx, ('sx',), site)
        if by != 0.0:
            H += (by, ('sy',), site)
        if bz != 0.0:
            H += (bz, ('sz',), site)

    return H
